/*---------------------------------------------------------------------------*/
/*        Copyright (c) 1996 LAL Orsay, UPS-IN2P3-CNRS (France).             */
/*                                                                           */
/* Redistribution and use in source and binary forms, with or without        */
/* modification, are permitted provided that the following conditions        */
/* are met:                                                                  */
/* 1. Redistributions of source code must retain the above copyright         */
/*    notice, this list of conditions and the following disclaimer.          */
/* 2. Redistributions in binary form must reproduce the above copyright      */
/*    notice, this list of conditions and the following disclaimer in the    */
/*    documentation and/or other materials provided with the distribution.   */
/* 3. All advertising materials mentioning features or use of this software  */
/*    must display the following acknowledgement:                            */
/*      This product includes software developed by the Computer Application */
/*      Development Group at LAL Orsay (Laboratoire de l'Accelerateur        */
/*      Linaire - UPS-IN2P3-CNRS).                                           */
/* 4. Neither the name of the Institute nor of the Laboratory may be used    */
/*    to endorse or promote products derived from this software without      */
/*    specific prior written permission.                                     */
/*                                                                           */
/* THIS SOFTWARE IS PROVIDED BY THE LAL AND CONTRIBUTORS ``AS IS'' AND       */
/* ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE     */
/* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR        */
/* PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE LAL OR CONTRIBUTORS BE      */
/* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR       */
/* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF      */
/* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS  */
/* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN   */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)   */
/* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF    */
/* THE POSSIBILITY OF SUCH DAMAGE.                                           */
/*---------------------------------------------------------------------------*/

/*- Header ------------------------------------------------------------*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <CmMessage.h>
#include <CSet.h>

#ifdef unix
#include <fcntl.h>
#endif

#include "Db.h"

/*---------------------------------------------------------------------*/
typedef struct
{
  DbError error;
  char error_text[256];
  void* object;
} TransactionInfo;
/*---------------------------------------------------------------------*/

/*---------------------------------------------------------------------*/
DbType   DbParserCreateType (char* name);
DbObject DbParserCreateObject (char* name);
DbType   DbParserLoadObjectVersion (char* name);

DbObject DbParserLoadObjectCopy (char* text);
/*---------------------------------------------------------------------*/

/*---- Message handlers -----------------------------------------------*/
#define DeclareMessageHandler(name) CmMessageStatus name (\
    CmMessage message, \
    char* sender, \
    char* server)


static DeclareMessageHandler (BreakHandler);
static DeclareMessageHandler (TypeDefinitionHandler);
static DeclareMessageHandler (TypeHeadersHandler);
static DeclareMessageHandler (ObjectDefinitionHandler);
static DeclareMessageHandler (ObjectDefsHandler);
static DeclareMessageHandler (ObjectNamesHandler);
static DeclareMessageHandler (TypeNamesHandler);
static DeclareMessageHandler (VersioningHandler);
static DeclareMessageHandler (VersionHistoryHandler);
static DeclareMessageHandler (ObjectHistoryHandler);

static DbType DbCreateType (char* definition);
static void   DisconnectionHandler (CmConnect connect);

/*---------------------------------------------------------------------*/

/*---------------------------------------------------------------------*/
static int    Initialized        = 0;
static int    Debug              = 0;
static int    TransactionPending = 0;

static char*  NamesBuffer        = NULL;
static int    NamesAllocated     = 0;
static int    NamesLength        = 0;

static DbVector* HistoryBuffer   = NULL;

static int    UseVersioning      = 0;

static CmConnectConnectionHandler PreviousDisconnectionHandler = NULL;
/*---------------------------------------------------------------------*/

/*---------------------------------------------------------------------*/
void DbStartup ()
/*---------------------------------------------------------------------*/
{
  if (!Initialized)
    {
      if (CmConnectWhoAmI () == NULL)
        {
          DbSetErrorInfo (DbErrorNoCm, "DbStartup>");
          return;
        }
      
      Initialized = 1;

      Debug = (getenv ("DBDEBUG") != NULL);

      PreviousDisconnectionHandler = NULL;
    }

  CmMessageInstallHandler (BreakHandler,             "DbBreak");
  CmMessageInstallHandler (TypeDefinitionHandler,    "DbTypeDefinition");
  CmMessageInstallHandler (TypeHeadersHandler,       "DbTypeHeaders");
  CmMessageInstallHandler (ObjectDefinitionHandler,  "DbObjectDefinition");
  
  CmMessageInstallHandler (ObjectDefsHandler,        "DbObjectDefs");
  
  CmMessageInstallHandler (ObjectNamesHandler,       "DbObjectNames");
  CmMessageInstallHandler (TypeNamesHandler,         "DbTypeNames");
  
  CmMessageInstallHandler (VersionHistoryHandler,    "DbVersionHistory");
  CmMessageInstallHandler (ObjectHistoryHandler,     "DbObjectHistory");
      
  CmMessageInstallHandler (VersioningHandler,        "DbVersioning");

  if (PreviousDisconnectionHandler == NULL)
    {
      PreviousDisconnectionHandler = CmConnectInstallDisconnectionHandler (
          DisconnectionHandler);
    }
}

/*---------------------------------------------------------------------*/
void DbCleanup ()
/*---------------------------------------------------------------------*/
{
  if (Initialized)
    {
      Initialized = 0;

        /*
      DbObjectCleanup ();
      DbTypeCleanup ();
      DbHandleCleanup ();
      
      GCCleanup ();
        */
    }
}

/*---------------------------------------------------------------------*/
void DbCreateObject (char* definition)
/*---------------------------------------------------------------------*/
{
  CmMessage message;
  
  DbStartup ();
  if (!Initialized) return;
  
  if ((definition == NULL) ||
      (strlen (definition) == 0)) 
    {
      DbSetErrorInfo (DbErrorBadArgument, "DbCreateObject>");
      return;
    }
  
  for (;;)
    {
      int transaction;
      TransactionInfo info = {DbErrorOk, "", NULL};

      DbClearErrorInfo ();

      transaction = CmOpenTransaction ("DbCreateObject", &info);

      message = CmMessageNew ();
      
      CmMessagePutText (message, definition);
      CmMessagePutInt  (message, transaction);

      CmMessageSetType (message, "DbCreateObject");
      
      if (CmMessageSend (message, DbServerName ()))
        {
          DbWaitForTransaction (transaction);
        }
      else
        {
          DbSetErrorInfo (DbErrorNoServer, "DbCreateObject>");
        }
      
      CmMessageDelete (message);

      CmCloseTransaction (transaction);

      DbSetErrorInfo (info.error, info.error_text);
      
      if (DbGetError () == DbErrorUnloadedStructure)
        {
          if (DbLoadType (DbGetErrorInfo ()) == NULL) break;
        }
      else
        {
          break;
        }
    }
}

/*---------------------------------------------------------------------*/
void DbCreateObjects (char* definition, ...)
/*---------------------------------------------------------------------*/
{
  CmMessage message;
  
  DbStartup ();
  if (!Initialized) return;
  
  if ((definition == NULL) ||
      (strlen (definition) == 0)) 
    {
      DbSetErrorInfo (DbErrorBadArgument, "DbCreateObjects>");
      return;
    }
  
  for (;;)
    {
      int transaction;
      TransactionInfo info = {DbErrorOk, "", NULL};

      DbClearErrorInfo ();
      
      transaction = CmOpenTransaction ("DbCreateObjects", &info);

      message = CmMessageNew ();
      
      CmMessagePutText (message, definition);
      CmMessagePutInt  (message, transaction);

      CmMessageSetType (message, "DbCreateObjects");
      
      if (CmMessageSend (message, DbServerName ()))
        {
          DbWaitForTransaction (transaction);
        }
      else
        {
          DbSetErrorInfo (DbErrorNoServer, "DbCreateObject>");
        }
      
      CmMessageDelete (message);
      
      CmCloseTransaction (transaction);
      DbSetErrorInfo (info.error, info.error_text);

      if (DbGetError () == DbErrorUnloadedStructure)
        {
          if (DbLoadType (DbGetErrorInfo ()) == NULL) break;
        }
      else
        {
          break;
        }
    }
}

/*---------------------------------------------------------------------*/
DbObject DbLoadObject (char* name)
/*---------------------------------------------------------------------*/
{
  DbObject   result = NULL;
  CmMessage  message;
  char       object_name[256];
  char       path_name[256];
  char*      slash;
  
  DbStartup ();
  if (!Initialized) return (NULL);
  
  if (name == NULL)
    {
      DbSetErrorInfo (DbErrorBadArgument, "DbLoadObject>");
      return (NULL);
    }
  
  slash = strrchr (name, '/');
  if (slash == NULL)
    {
      strcpy (object_name, name);
      path_name[0] = 0;
    }
  else
    {
      strcpy (path_name, name);
      slash = strrchr (path_name, '/');
      *slash = 0;
      strcpy (object_name, slash + 1);
    }
  
  if (strlen (object_name) == 0)
    {
      DbSetErrorInfo (DbErrorBadArgument, "DbLoadObject>");
      return (NULL);
    }
  
  for (;;)
    {
      int transaction;
      TransactionInfo info = {DbErrorOk, "", NULL};

      DbClearErrorInfo ();
      
      transaction = CmOpenTransaction ("DbLoadObject", &info);

      message = CmMessageNew ();
      
      CmMessagePutText (message, object_name);
      CmMessagePutText (message, path_name);
      CmMessagePutInt  (message, transaction);

      CmMessageSetType (message, "DbLoadObject");
      
      if (CmMessageSend (message, DbServerName ()))
        {
          DbWaitForTransaction (transaction);
          result = DbObjectFindInPath (object_name, path_name);
        }
      else
        {
          DbSetErrorInfo (DbErrorNoServer, "DbLoadObject>");
        }
      
      CmMessageDelete (message);
      GCCleanup ();

      CmCloseTransaction (transaction);
      DbSetErrorInfo (info.error, info.error_text);

      if (DbGetError () == DbErrorUnloadedStructure)
        {
          if (DbLoadType (DbGetErrorInfo ()) == NULL) break;
        }
      else
        {
          break;
        }
    }

  return (result);
}

/*---------------------------------------------------------------------*/
void DbLoadObjects (char* pattern, CLList list)
/*---------------------------------------------------------------------*/
{
  CmMessage  message;
  char       name_pattern[256];
  char       path_pattern[256];
  char*      slash;
  
  DbStartup ();
  if (!Initialized) return;

      /*
        On suppose que une seule expression pattern. est utilisee
        (plutot qu'une liste d'expressions)
        */
  
  if (pattern == NULL)
    {
      DbSetErrorInfo (DbErrorBadArgument, "DbLoadObjects>");
      return;
    }
  
  slash = strrchr (pattern, '/');
  if (slash == NULL)
    {
      strcpy (name_pattern, pattern);
      path_pattern[0] = 0;
    }
  else
    {
      strcpy (path_pattern, pattern);
      slash = strrchr (path_pattern, '/');
      *slash = 0;
      strcpy (name_pattern, slash + 1);
    }
  
  if (strlen (name_pattern) == 0)
    {
      DbSetErrorInfo (DbErrorBadArgument, "DbLoadObjects>");
      return;
    }

  for (;;)
    {
      int transaction;
      TransactionInfo info = {DbErrorOk, "", NULL};

      DbClearErrorInfo ();

      info.object = list;

      transaction = CmOpenTransaction ("DbLoadObjects", &info);

      message = CmMessageNew ();
      
      CmMessagePutText (message, name_pattern);
      CmMessagePutText (message, path_pattern);
      CmMessagePutInt  (message, transaction);

      CmMessageSetType (message, "DbLoadObjects");
      
      if (CmMessageSend (message, DbServerName ()))
        {
          DbWaitForTransaction (transaction);
        }
      else
        {
          DbSetErrorInfo (DbErrorNoServer, "DbLoadObjects>");
        }
      
      CmMessageDelete (message);

      CmCloseTransaction (transaction);
      DbSetErrorInfo (info.error, info.error_text);

      if (DbGetError () == DbErrorUnloadedStructure)
        {
          if (DbLoadType (DbGetErrorInfo ()) == NULL) break;
        }
      else
        {
          break;
        }
    }

  if (DbDebug ())
    {
      printf ("DbLoadObjects> List size : %d\n", CLListCount (list));
    }
}

/*---------------------------------------------------------------------*/
DbObject DbLoadObjectByTag (char* name, char* tag)
/*---------------------------------------------------------------------*/
{
  DbObject   result = NULL;
  CmMessage  message;
  char       object_name[256];
  char       path_name[256];
  char*      slash;
 
  DbStartup ();
  if (!Initialized) return (NULL);
  
  if (name == NULL || tag == NULL)
    {
      DbSetErrorInfo (DbErrorBadArgument, "DbLoadObjectByTag>");
      return (NULL);
    }
  
  slash = strrchr (name, '/');
  if (slash == NULL)
    {
      strcpy (object_name, name);
      path_name[0] = 0;
    }
  else
    {
      strcpy (path_name, name);
      slash = strrchr (path_name, '/');
      *slash = 0;
      strcpy (object_name, slash + 1);
    }
  
  if (strlen (object_name) == 0)
    {
      DbSetErrorInfo (DbErrorBadArgument, "DbLoadObjectByTag>");
      return (NULL);
    }
  
  for (;;)
    {
      int transaction;
      TransactionInfo info = {DbErrorOk, "", NULL};

      DbClearErrorInfo ();
      
      transaction = CmOpenTransaction ("DbLoadObjectByTag", &info);

      message = CmMessageNew ();
      
      CmMessagePutText (message, object_name);
      CmMessagePutText (message, path_name);
      CmMessagePutText (message, tag);
      CmMessagePutInt  (message, transaction);

      CmMessageSetType (message, "DbLoadObjByTag");
      
      if (CmMessageSend (message, DbServerName ()))
        {
          DbWaitForTransaction (transaction);
          result = DbObjectFindInPath (object_name, path_name);
        }
      else
        {
          DbSetErrorInfo (DbErrorNoServer, "DbLoadObjectByTag>");
        }
      
      CmMessageDelete (message);
      
      CmCloseTransaction (transaction);
      DbSetErrorInfo (info.error, info.error_text);

      if (DbGetError () == DbErrorUnloadedStructure)
        {
          if (DbLoadType (DbGetErrorInfo ()) == NULL) break;
        }
      else
        {
          break;
        }
    }
  
  return (result);  
}

/*---------------------------------------------------------------------*/
DbObject DbLoadObjectByDate (char* name, char* date)
/*---------------------------------------------------------------------*/
{
  DbObject   result = NULL;
  CmMessage  message;
  char       object_name[256];
  char       path_name[256];
  char*      slash;
  
  DbStartup ();
  if (!Initialized) return (NULL);
  
  if (name == NULL || date == NULL)
    {
      DbSetErrorInfo (DbErrorBadArgument, "DbLoadObjectByDate");
      return (NULL);
    }
  
  slash = strrchr (name, '/');
  if (slash == NULL)
    {
      strcpy (object_name, name);
      path_name[0] = 0;
    }
  else
    {
      strcpy (path_name, name);
      slash = strrchr (path_name, '/');
      *slash = 0;
      strcpy (object_name, slash + 1);
    }
  
  if (strlen (object_name) == 0)
    {
      DbSetErrorInfo (DbErrorBadArgument, "DbLoadObjByDate>");
      return (NULL);
    }
  
  for (;;)
    {
      int transaction;
      TransactionInfo info = {DbErrorOk, "", NULL};

      DbClearErrorInfo ();
      
      transaction = CmOpenTransaction ("DbLoadObjectByDate", &info);

      message = CmMessageNew ();
      
      CmMessagePutText (message, object_name);
      CmMessagePutText (message, path_name);
      CmMessagePutText (message, date);
      CmMessagePutInt  (message, transaction);

      CmMessageSetType (message, "DbLoadObjByDate");
      
      if (CmMessageSend (message, DbServerName ()))
        {
          DbWaitForTransaction (transaction);
          result = DbObjectFindInPath (object_name, path_name);
        }
      else
        {
          DbSetErrorInfo (DbErrorNoServer, "DbLoadObjectByDate>");
        }
      
      CmMessageDelete (message);

      CmCloseTransaction (transaction);
      DbSetErrorInfo (info.error, info.error_text);
      
      if (DbGetError () == DbErrorUnloadedStructure)
        {
          if (DbLoadType (DbGetErrorInfo ()) == NULL) break;
        }
      else
        {
          break;
        }
    }
  
  return (result);  
}

/*---------------------------------------------------------------------*/
DbObject DbLoadObjectFromBranch (char* name, char* branch)
/*---------------------------------------------------------------------*/
{
  DbObject   result = NULL;
  CmMessage  message;
  char       object_name[256];
  char       path_name[256];
  char*      slash;
 
  DbStartup ();
  if (!Initialized) return (NULL);
  
  if (name == NULL || branch == NULL)
    {
      DbSetErrorInfo (DbErrorBadArgument, "DbLoadObjectFromBranch>");
      return (NULL);
    }
  
  slash = strrchr (name, '/');
  if (slash == NULL)
    {
      strcpy (object_name, name);
      path_name[0] = 0;
    }
  else
    {
      strcpy (path_name, name);
      slash = strrchr (path_name, '/');
      *slash = 0;
      strcpy (object_name, slash + 1);
    }
  
  if (strlen (object_name) == 0)
    {
      DbSetErrorInfo (DbErrorBadArgument, "DbLoadObjectFromBranch>");
      return (NULL);
    }
  
  for (;;)
    {
      int transaction;
      TransactionInfo info = {DbErrorOk, "", NULL};

      DbClearErrorInfo ();
      
      transaction = CmOpenTransaction ("DbLoadObjectFromBranch", &info);

      message = CmMessageNew ();
      
      CmMessagePutText (message, object_name);
      CmMessagePutText (message, path_name);
      CmMessagePutText (message, branch);
      CmMessagePutInt  (message, transaction);

      CmMessageSetType (message, "DbLoadObjFromBranch");
      
      if (CmMessageSend (message, DbServerName ()))
        {
          DbWaitForTransaction (transaction);
          result = DbObjectFindInPath (object_name, path_name);
        }
      else
        {
          DbSetErrorInfo (DbErrorNoServer, "DbLoadObjectFromBranch>");
        }
      
      CmMessageDelete (message);
      
      CmCloseTransaction (transaction);
      DbSetErrorInfo (info.error, info.error_text);
      
      if (DbGetError () == DbErrorUnloadedStructure)
        {
          if (DbLoadType (DbGetErrorInfo ()) == NULL) break;
        }
      else
        {
          break;
        }
    }
  
  return (result);  
}

/*---------------------------------------------------------------------*/
void DbLoadObjectsByTag (char* pattern, char* tag, CLList list)
/*---------------------------------------------------------------------*/
{
  CmMessage  message;
  char       name_pattern[256];
  char       path_pattern[256];
  char*      slash;
 
  DbStartup ();
  if (!Initialized) return;
  
      /*
        On suppose, la encore, qu'une seule expression pattern. est
        utilisee (plutot qu'une liste d'expressions)
      */
  
  if (pattern == NULL || tag == NULL)
    {
      DbSetErrorInfo (DbErrorBadArgument, "DbLoadObjectsByTag>");
      return;
    }
  
  slash = strrchr (pattern, '/');
  if (slash == NULL)
    {
      strcpy (name_pattern, pattern);
      path_pattern[0] = 0;
    }
  else
    {
      strcpy (path_pattern, pattern);
      slash = strrchr (path_pattern, '/');
      *slash = 0;
      strcpy (name_pattern, slash + 1);
    }
  
  if (strlen (name_pattern) == 0)
    {
      DbSetErrorInfo (DbErrorBadArgument, "DbLoadObjectsByTag>");
      return;
    }
  
  for (;;)
    {
      int transaction;
      TransactionInfo info = {DbErrorOk, "", NULL};

      DbClearErrorInfo ();
      
      info.object = list;
      transaction = CmOpenTransaction ("DbLoadObjsByTag", &info);

      message = CmMessageNew ();
      
      CmMessagePutText (message, name_pattern);
      CmMessagePutText (message, path_pattern);
      CmMessagePutText (message, tag);
      CmMessagePutInt  (message, transaction);

      CmMessageSetType (message, "DbLoadObjsByTag");
      
      if (CmMessageSend (message, DbServerName ()))
        {
          DbWaitForTransaction (transaction);
        }
      else
        {
          DbSetErrorInfo (DbErrorNoServer, "DbLoadObjectsByTag>");
        }
      
      CmMessageDelete (message);
      
      CmCloseTransaction (transaction);
      DbSetErrorInfo (info.error, info.error_text);
      
      if (DbGetError () == DbErrorUnloadedStructure)
        {
          if (DbLoadType (DbGetErrorInfo ()) == NULL) break;
        }
      else
        {
          break;
        }
    }
  
  if (DbDebug ())
    {
      printf ("DbLoadObjectsByTag> List size : %d\n", CLListCount (list));
    }
}

/*---------------------------------------------------------------------*/
void DbLoadObjectsByDate (char* pattern, char* date, CLList list)
/*---------------------------------------------------------------------*/
{
  CmMessage  message;
  char       name_pattern[256];
  char       path_pattern[256];
  char*      slash;
  
  DbStartup ();
  if (!Initialized) return;
  
      /*
        On suppose, la encore, qu'une seule expression pattern. est
        utilisee (plutot qu'une liste d'expressions)
      */
  
  if (pattern == NULL || date == NULL)
    {
      DbSetErrorInfo (DbErrorBadArgument, "DbLoadObjectsByDate");
      return;
    }
  
  slash = strrchr (pattern, '/');
  if (slash == NULL)
    {
      strcpy (name_pattern, pattern);
      path_pattern[0] = 0;
    }
  else
    {
      strcpy (path_pattern, pattern);
      slash = strrchr (path_pattern, '/');
      *slash = 0;
      strcpy (name_pattern, slash + 1);
    }
  
  if (strlen (name_pattern) == 0)
    {
      DbSetErrorInfo (DbErrorBadArgument, "DbLoadObjsByDate>");
      return;
    }
  
  for (;;)
    {
      int transaction;
      TransactionInfo info = {DbErrorOk, "", NULL};

      DbClearErrorInfo ();
      
      info.object = list;
      transaction = CmOpenTransaction ("DbLoadObjsByDate", &info);

      message = CmMessageNew ();
      
      CmMessagePutText (message, name_pattern);
      CmMessagePutText (message, path_pattern);
      CmMessagePutText (message, date);
      CmMessagePutInt  (message, transaction);

      CmMessageSetType (message, "DbLoadObjsByDate");
      
      if (CmMessageSend (message, DbServerName ()))
        {
          DbWaitForTransaction (transaction);
        }
      else
        {
          DbSetErrorInfo (DbErrorNoServer, "DbLoadObjectsByDate>");
        }
      
      CmMessageDelete (message);

      CmCloseTransaction (transaction);
      DbSetErrorInfo (info.error, info.error_text);
      
      if (DbGetError () == DbErrorUnloadedStructure)
        {
          if (DbLoadType (DbGetErrorInfo ()) == NULL) break;
        }
      else
        {
          break;
        }
    }
  
  if (DbDebug ())
    {
      printf ("DbLoadObjectsByDate> List size : %d\n", CLListCount (list));
    }
}

/*---------------------------------------------------------------------*/
void DbTagObject (char* name, char* tag)
/*---------------------------------------------------------------------*/
{
  CmMessage  message;
  char       object_name[256];
  char       path_name[256];
  char*      slash;
  int        transaction;
  TransactionInfo info = {DbErrorOk, "", NULL};
  
  DbStartup ();
  if (!Initialized) return;
  
  if (name == NULL || tag == NULL)
    {
      DbSetErrorInfo (DbErrorBadArgument, "DbTagObject");
      return;
    }
  
  slash = strrchr (name, '/');
  if (slash == NULL)
    {
      strcpy (object_name, name);
      path_name[0] = 0;
    }
  else
    {
      strcpy (path_name, name);
      slash = strrchr (path_name, '/');
      *slash = 0;
      strcpy (object_name, slash + 1);
    }
  
  if (strlen (object_name) == 0)
    {
      DbSetErrorInfo (DbErrorBadArgument, "DbTagObject>");
      return;
    }
  
  DbClearErrorInfo ();
      
  transaction = CmOpenTransaction ("DbTagObject", &info);

  message = CmMessageNew ();
      
  CmMessagePutText (message, object_name);
  CmMessagePutText (message, path_name);
  CmMessagePutText (message, tag);
  CmMessagePutInt  (message, transaction);
  
  CmMessageSetType (message, "DbTagObject");
      
  if (CmMessageSend (message, DbServerName ()))
    {
      DbWaitForTransaction (transaction);
    }
  else
    {
      DbSetErrorInfo (DbErrorNoServer, "DbTagObject>");
    }
  
  CmMessageDelete (message);
  
  CmCloseTransaction (transaction);
  DbSetErrorInfo (info.error, info.error_text);
      
  return;  
}

/*---------------------------------------------------------------------*/
void DbTagObjectsInPath (char* path_name, char* tag)
/*---------------------------------------------------------------------*/
{
  CmMessage  message;

  DbStartup ();
  if (!Initialized) return;
  
  if (path_name == NULL || tag == NULL)
    {
      DbSetErrorInfo (DbErrorBadArgument, "DbTagObjectsInPath");
      return;
    }
  
  for (;;)
    {
      int transaction;
      TransactionInfo info = {DbErrorOk, "", NULL};

      DbClearErrorInfo ();
      
      transaction = CmOpenTransaction ("DbTagObjectInPath", &info);

      message = CmMessageNew ();
      
      CmMessagePutText (message, path_name);
      CmMessagePutText (message, tag);
      CmMessagePutInt  (message, transaction);

      CmMessageSetType (message, "DbTagObjectsInPath");
      
      if (CmMessageSend (message, DbServerName ()))
        {
          DbWaitForTransaction (transaction);
        }
      else
        {
          DbSetErrorInfo (DbErrorNoServer, "DbTagObjectInPath>");
        }
      
      CmMessageDelete (message);

      CmCloseTransaction (transaction);
      DbSetErrorInfo (info.error, info.error_text);
      
      if (DbGetError () == DbErrorUnloadedStructure)
        {
          if (DbLoadType (DbGetErrorInfo ()) == NULL) break;
        }
      else
        {
          break;
        }
    }
  
  return;  
}

/*---------------------------------------------------------------------*/
void DbTagDatabase (char* tag)
/*---------------------------------------------------------------------*/
{
  CmMessage  message;
  
  DbStartup ();
  if (!Initialized) return;
  
  if (tag == NULL)
    {
      DbSetErrorInfo (DbErrorBadArgument, "DbTagDatabase");
      return;
    }
  
  for (;;)
    {
      int transaction;
      TransactionInfo info = {DbErrorOk, "", NULL};

      DbClearErrorInfo ();
      
      transaction = CmOpenTransaction ("DbTagDatabase", &info);

      message = CmMessageNew ();
      
      CmMessagePutText (message, tag);
      CmMessagePutInt  (message, transaction);

      CmMessageSetType (message, "DbTagDatabase");
      
      if (CmMessageSend (message, DbServerName ()))
        {
          DbWaitForTransaction (transaction);
        }
      else
        {
          DbSetErrorInfo (DbErrorNoServer, "DbTagDatabase>");
        }
      
      CmMessageDelete (message);

      CmCloseTransaction (transaction);
      DbSetErrorInfo (info.error, info.error_text);
      
      if (DbGetError () == DbErrorUnloadedStructure)
        {
          if (DbLoadType (DbGetErrorInfo ()) == NULL) break;
        }
      else
        {
          break;
        }
    }
  
  return;  
}

/*---------------------------------------------------------------------*/
void DbDeleteObjectTag (char* name, char* tag)
/*---------------------------------------------------------------------*/
{
  CmMessage  message;
  char       object_name[256];
  char       path_name[256];
  char*      slash;
  
  DbStartup ();
  if (!Initialized) return;
  
  if (name == NULL || tag == NULL)
    {
      DbSetErrorInfo (DbErrorBadArgument, "DbDeleteObjectTag");
      return;
    }
  
  slash = strrchr (name, '/');
  if (slash == NULL)
    {
      strcpy (object_name, name);
      path_name[0] = 0;
    }
  else
    {
      strcpy (path_name, name);
      slash = strrchr (path_name, '/');
      *slash = 0;
      strcpy (object_name, slash + 1);
    }
  
  if (strlen (object_name) == 0)
    {
      DbSetErrorInfo (DbErrorBadArgument, "DbDeleteObjectTag>");
      return;
    }
  
  for (;;)
    {
      int transaction;
      TransactionInfo info = {DbErrorOk, "", NULL};

      DbClearErrorInfo ();
      
      transaction = CmOpenTransaction ("DbDeleteObjectTag", &info);

      message = CmMessageNew ();
      
      CmMessagePutText (message, object_name);
      CmMessagePutText (message, path_name);
      CmMessagePutText (message, tag);
      CmMessagePutInt  (message, transaction);

      CmMessageSetType (message, "DbDeleteObjectTag");
      
      if (CmMessageSend (message, DbServerName ()))
        {
          DbWaitForTransaction (transaction);
        }
      else
        {
          DbSetErrorInfo (DbErrorNoServer, "DbDeleteObjectTag>");
        }
      
      CmMessageDelete (message);

      CmCloseTransaction (transaction);
      DbSetErrorInfo (info.error, info.error_text);
      
      if (DbGetError () == DbErrorUnloadedStructure)
        {
          if (DbLoadType (DbGetErrorInfo ()) == NULL) break;
        }
      else
        {
          break;
        }
    }
  
  return;  
}


/*---------------------------------------------------------------------*/
void DbCreateBranch (char* tag, char* branch)
/*---------------------------------------------------------------------*/
{
  CmMessage  message;
  char       object_name[256];
  char       path_name[256];
  char*      slash;
  
  DbStartup ();
  if (!Initialized) return;
  
  if (tag == NULL || branch == NULL)
    {
      DbSetErrorInfo (DbErrorBadArgument, "DbCreateBranch");
      return;
    }
  
  for (;;)
    {
      int transaction;
      TransactionInfo info = {DbErrorOk, "", NULL};

      DbClearErrorInfo ();
      
      transaction = CmOpenTransaction ("DbCreateBranch", &info);

      message = CmMessageNew ();
      
      CmMessagePutText (message, tag);
      CmMessagePutText (message, branch);
      CmMessagePutInt  (message, transaction);

      CmMessageSetType (message, "DbCreateBranch");
      
      if (CmMessageSend (message, DbServerName ()))
        {
          DbWaitForTransaction (transaction);
        }
      else
        {
          DbSetErrorInfo (DbErrorNoServer, "DbCreateBranch>");
        }
      
      CmMessageDelete (message);

      CmCloseTransaction (transaction);
      DbSetErrorInfo (info.error, info.error_text);
      
      if (DbGetError () == DbErrorUnloadedStructure)
        {
          if (DbLoadType (DbGetErrorInfo ()) == NULL) break;
        }
      else
        {
          break;
        }
    }
  
  return;  
}


/*---------------------------------------------------------------------*/
void DbCommitObjectInBranch (char* name, char* btag)
/*---------------------------------------------------------------------*/
{
  DbObject object;
  char     object_name[256];
  char     path_name[256];
  char*    slash;
  
  DbStartup ();
  if (!Initialized) return;
  
  DbClearErrorInfo ();
  
  if ((name == NULL) || (strlen (name) == 0)
                     ||
      (btag == NULL) || (strlen(btag) == 0)) 
  {
    DbSetErrorInfo (DbErrorBadArgument, "DbCommitObjectInBranch>");
    return;
  }
  
  slash = strrchr (name, '/');
  if (slash == NULL)
  {
    strcpy (object_name, name);
    path_name[0] = 0;
  }
  else
  {
    strcpy (path_name, name);
    slash = strrchr (path_name, '/');
    *slash = 0;
    strcpy (object_name, slash + 1);
  }
  
  if (strlen (object_name) == 0)
  {
    DbSetErrorInfo (DbErrorBadArgument, "DbCommitObjectInBranch>");
    return;
  }

  object = DbObjectFindInPath (object_name, path_name);

  DbObjectCommitInBranch (object, btag);
}

/*---------------------------------------------------------------------*/
void DbObjectCommitInBranch (DbObject object, char* btag)
/*---------------------------------------------------------------------*/
{
  CmMessage  message;
  char*      text;

  if ((object == NULL) || (btag == NULL) || (strlen(btag) == 0))
    {
      DbSetErrorInfo (DbErrorBadArgument, "DbObjectCommitInBranch>");
      return;
    }
  
  DbStartup ();
  if (!Initialized) return;
  
  DbClearErrorInfo ();
  
  for (;;)
    {
      int transaction;
      TransactionInfo info = {DbErrorOk, "", NULL};

      DbClearErrorInfo ();
      
      transaction = CmOpenTransaction ("DbObjectCommitInBranch", &info);

      message = CmMessageNew ();

      text = DbConvertObjectToString (object);
      CmMessagePutText (message, text);
      CmMessagePutText (message, btag);
      CmMessagePutInt  (message, transaction);

      CmMessageSetType (message, "DbCiObjInBranch");
      
      if (CmMessageSend (message, DbServerName ()))
        {
          DbWaitForTransaction (transaction);
        }
      else
        {
          DbSetErrorInfo (DbErrorNoServer, "DbObjectCommitInBranch>");
        }
  
      CmMessageDelete (message);

      CmCloseTransaction (transaction);
      DbSetErrorInfo (info.error, info.error_text);

      if (DbGetError () == DbErrorUnloadedStructure)
        {
          if (DbLoadType (DbGetErrorInfo ()) == NULL) break;
        }
      else
        {
          break;
        }
    }
  return;
}

/*---------------------------------------------------------------------*/
DbVector* DbGetVersionHistory (char* object_name)
/*---------------------------------------------------------------------*/
{
  char* list;
  CmMessage message;
  int transaction;
  TransactionInfo info = {DbErrorOk, "", NULL};

  DbStartup ();
  if (!Initialized) return (NULL);
  
  if (object_name == NULL)
    {
      DbSetErrorInfo (DbErrorBadArgument, "DbGetVersionHistory>");
    }
  
  if (DbDebug ())
    {
      DbPrint ("DbGetVersionHistory>  %s\n", object_name);
    }
  
  if (HistoryBuffer == NULL)
    {
      HistoryBuffer = DbStringVectorNew ();
    }
  else
    {
      DbVectorClear (HistoryBuffer);
    }
  
  DbClearErrorInfo ();
  
  transaction = CmOpenTransaction ("DbGetVersionHistory", &info);

  message = CmMessageNew ();
  
  CmMessagePutText (message, object_name);
  CmMessagePutInt  (message, transaction);

  CmMessageSetType (message, "DbGetVersionHistory");
  
  if (CmMessageSend (message, DbServerName ()))
    {
      DbWaitForTransaction (transaction);
    }
  else
    {
      DbSetErrorInfo (DbErrorNoServer, "DbGetVersionHistory>");
    }
  
  CmMessageDelete (message);

  CmCloseTransaction (transaction);
  DbSetErrorInfo (info.error, info.error_text);
  
  if (DbDebug ())
    {
      int last;
      int index;

      last = DbVectorGetSize (HistoryBuffer);

      for (index = 0; index < last; index++)
        {
          DbString* string;

          string = DbStringVectorGetElement (HistoryBuffer, index);
          printf ("DbGetVersionHistory> HistoryBuffer : %s\n", 
                  DbStringGetContents (string));
        }
    }
  
  return (HistoryBuffer);
}

/*---------------------------------------------------------------------*/
DbVector* DbGetObjectHistory (DbObject object,
                              char* from_date, char* to_date)
/*---------------------------------------------------------------------*/
{
  DbVector*  result = NULL;
  CmMessage  message;
  char*      object_name;
  char*      path_name;
  char*      slash;
  
  DbStartup ();
  if (!Initialized) return (NULL);
  
  if (object == NULL)
    {
      DbSetErrorInfo (DbErrorUnknownObject, "DbGetObjectHistory");
      return (NULL);
    }
  
  if (from_date == NULL || to_date == NULL)
    {
      DbSetErrorInfo (DbErrorBadArgument, "DbGetObjectHistory");
      return (NULL);
    }
  
  object_name = DbObjectGetName (object);
  path_name   = DbObjectGetPathName (object);
  
  if (strlen (object_name) == 0)
    {
      DbSetErrorInfo (DbErrorBadArgument, "DbGetObjectHistory>");
      return (NULL);
    }
  
  DbClearErrorInfo ();

  DbSetErrorInfo (DbErrorNotYetImplemented, "DbGetObjectHistory>");
      
  return (result);
}

/*---------------------------------------------------------------------*/
void DbObjectCommit (DbObject object)
/*---------------------------------------------------------------------*/
{
  CmMessage  message;
  char* text;
  int transaction;
  TransactionInfo info = {DbErrorOk, "", NULL};

  if (object == NULL) return;

  transaction = CmOpenTransaction ("DbCommitObject", &info);

  message = CmMessageNew ();
  
  text = DbConvertObjectToString (object);
  CmMessagePutText (message, text);
  CmMessagePutInt  (message, transaction);

  CmMessageSetType (message, "DbCommitObject");
  
  if (CmMessageSend (message, DbServerName ()))
    {
      DbWaitForTransaction (transaction);
    }
  else
    {
      DbSetErrorInfo (DbErrorNoServer, "DbObjectCommit>");
    }
  
  CmMessageDelete (message);

  CmCloseTransaction (transaction);
  DbSetErrorInfo (info.error, info.error_text);
}

/*---------------------------------------------------------------------*/
void DbObjectCommitAndHoldLock (DbObject object)
/*---------------------------------------------------------------------*/
{
  CmMessage  message;
  char* text;
  int transaction;
  TransactionInfo info = {DbErrorOk, "", NULL};

  if (object == NULL) return;

  transaction = CmOpenTransaction ("DbCommitObjectAndHoldLock", &info);

  message = CmMessageNew ();
  
  text = DbConvertObjectToString (object);
  CmMessagePutText (message, text);
  CmMessagePutInt  (message, transaction);

  CmMessageSetType (message, "DbCommitObjectAndHoldLock");
  
  if (CmMessageSend (message, DbServerName ()))
    {
      DbWaitForTransaction (transaction);
    }
  else
    {
      DbSetErrorInfo (DbErrorNoServer, "DbCommitObjectAndHoldLock>");
    }
  
  CmMessageDelete (message);

  CmCloseTransaction (transaction);
  DbSetErrorInfo (info.error, info.error_text);
}

/*---------------------------------------------------------------------*/
void DbCommitObject (char* name)
/*---------------------------------------------------------------------*/
{
  DbObject object;
  char object_name[256];
  char path_name[256];
  char* slash;
  
  DbStartup ();
  if (!Initialized) return;
  
  DbClearErrorInfo ();
  
  if ((name == NULL) || (strlen (name) == 0)) 
    {
      DbSetErrorInfo (DbErrorBadArgument, "DbCommitObject>");
      return;
    }
  
  slash = strrchr (name, '/');
  if (slash == NULL)
    {
      strcpy (object_name, name);
      path_name[0] = 0;
    }
  else
    {
      strcpy (path_name, name);
      slash = strrchr (path_name, '/');
      *slash = 0;
      strcpy (object_name, slash + 1);
    }
  
  if (strlen (object_name) == 0)
    {
      DbSetErrorInfo (DbErrorBadArgument, "DbCommitObject>");
      return;
    }

  object = DbObjectFindInPath (object_name, path_name);

  DbObjectCommit (object);
}


/*---------------------------------------------------------------------*/
void DbCommitObjects (char* pattern)
/*---------------------------------------------------------------------*/
{
  CmMessage  message;
  char*      text;
  char       object_name[256];
  char       path_name[256];

  int        transaction;
  TransactionInfo info = {DbErrorOk, "", NULL};
  
  if (pattern == NULL)
    {
      DbSetErrorInfo (DbErrorBadArgument, "DbCommitObjects>");
      return;
    }

      /* liste des noms des objets */
  text = DbObjectGetNames (pattern);

  transaction = CmOpenTransaction ("DbCiObjects", &info);

  message = CmMessageNew ();

      /* analyse de la liste des objets */
  while (strlen (text) > 0)
  {
    DbObject object = NULL;
    char* name;
    char* blank;
    char* slash;

    blank = strchr (text, ' ');
    if (blank != NULL)
    {
      name = text;
      text = blank + 1;
      *blank = 0;
    }
    else
    {
      name = text;
    }
  
    slash = strrchr (name, '/');
    if (slash == NULL)
    {
      strcpy (object_name, name);
      path_name[0] = 0;
    }
    else
    {
      strcpy (path_name, name);
      slash = strrchr (path_name, '/');
      *slash = 0;
      strcpy (object_name, slash + 1);
    }

    object = DbObjectFindInPath (object_name, path_name);

    if (object != NULL)
    {
      char* text;
      
      text = DbConvertObjectToString (object);
      CmMessagePutText (message, text);
    }
    
    if (blank == NULL) break;
  }
  
  CmMessagePutInt  (message, transaction);
  CmMessageSetType (message, "DbCiObjects");
  
  if (CmMessageSend (message, DbServerName ()))
    {
      DbWaitForTransaction (transaction);
    }
  else
    {
      DbSetErrorInfo (DbErrorNoServer, "DbCommitObjects>");
    }
  
  CmMessageDelete (message);

  CmCloseTransaction (transaction);
  DbSetErrorInfo (info.error, info.error_text);
}


/*---------------------------------------------------------------------*/
void DbCommitObjectAndHoldLock (char* name)
/*---------------------------------------------------------------------*/
{
  DbObject object;
  char object_name[256];
  char path_name[256];
  char* slash;
  
  DbStartup ();
  if (!Initialized) return;
  
  DbClearErrorInfo ();
  
  if ((name == NULL) || (strlen (name) == 0)) 
    {
      DbSetErrorInfo (DbErrorBadArgument, "DbCommitObjectAndHoldLock>");
      return;
    }
  
  slash = strrchr (name, '/');
  if (slash == NULL)
    {
      strcpy (object_name, name);
      path_name[0] = 0;
    }
  else
    {
      strcpy (path_name, name);
      slash = strrchr (path_name, '/');
      *slash = 0;
      strcpy (object_name, slash + 1);
    }
  
  if (strlen (object_name) == 0)
    {
      DbSetErrorInfo (DbErrorBadArgument, "DbCommitObjectAndHoldLock>");
      return;
    }

  object = DbObjectFindInPath (object_name, path_name);

  DbObjectCommitAndHoldLock (object);
}

/*---------------------------------------------------------------------*/
void DbLockObject (char* name)
/*---------------------------------------------------------------------*/
{
  char object_name[256];
  char path_name[256];
  char* slash;
  CmMessage  message;
  int transaction;
  TransactionInfo info = {DbErrorOk, "", NULL};

  DbStartup ();
  if (!Initialized) return;
  
  DbClearErrorInfo ();
  
  if ((name == NULL) || (strlen (name) == 0)) 
    {
      DbSetErrorInfo (DbErrorBadArgument, "DbLockObject>");
      return;
    }
  
  slash = strrchr (name, '/');
  if (slash == NULL)
    {
      strcpy (object_name, name);
      path_name[0] = 0;
    }
  else
    {
      strcpy (path_name, name);
      slash = strrchr (path_name, '/');
      *slash = 0;
      strcpy (object_name, slash + 1);
    }
  
  if (strlen (object_name) == 0)
    {
      DbSetErrorInfo (DbErrorBadArgument, "DbLockObject>");
      return;
    }

  transaction = CmOpenTransaction ("DbLockObject", &info);

  message = CmMessageNew ();
  
  CmMessagePutText (message, object_name);
  CmMessagePutText (message, path_name);
  CmMessagePutInt  (message, transaction);

  CmMessageSetType (message, "DbLockObject");
  
  if (CmMessageSend (message, DbServerName ()))
    {
      DbWaitForTransaction (transaction);
    }
  else
    {
      DbSetErrorInfo (DbErrorNoServer, "DbLockObject>");
    }
  
  CmMessageDelete (message);

  CmCloseTransaction (transaction);
  DbSetErrorInfo (info.error, info.error_text);
}

/*---------------------------------------------------------------------*/
void DbUnlockObject (char* name)
/*---------------------------------------------------------------------*/
{
  char object_name[256];
  char path_name[256];
  char* slash;
  CmMessage  message;
  int transaction;
  TransactionInfo info = {DbErrorOk, "", NULL};

  DbStartup ();
  if (!Initialized) return;
  
  DbClearErrorInfo ();
  
  if ((name == NULL) || (strlen (name) == 0)) 
    {
      DbSetErrorInfo (DbErrorBadArgument, "DbUnlockObject>");
      return;
    }
  
  slash = strrchr (name, '/');
  if (slash == NULL)
    {
      strcpy (object_name, name);
      path_name[0] = 0;
    }
  else
    {
      strcpy (path_name, name);
      slash = strrchr (path_name, '/');
      *slash = 0;
      strcpy (object_name, slash + 1);
    }
  
  if (strlen (object_name) == 0)
    {
      DbSetErrorInfo (DbErrorBadArgument, "DbUnlockObject>");
      return;
    }

  transaction = CmOpenTransaction ("DbUnlockObject", &info);

  message = CmMessageNew ();
  
  CmMessagePutText (message, object_name);
  CmMessagePutText (message, path_name);
  CmMessagePutInt  (message, transaction);

  CmMessageSetType (message, "DbUnlockObject");
  
  if (CmMessageSend (message, DbServerName ()))
    {
      DbWaitForTransaction (transaction);
    }
  else
    {
      DbSetErrorInfo (DbErrorNoServer, "DbUnlockObject>");
    }
  
  CmMessageDelete (message);

  CmCloseTransaction (transaction);
  DbSetErrorInfo (info.error, info.error_text);
}

/*---------------------------------------------------------------------*/
void DbAddTrigger (char* name, DbTriggerType trigger_type)
/*---------------------------------------------------------------------*/
{
  char object_name[256];
  char path_name[256];
  char* slash;
  CmMessage  message;
  int transaction;
  TransactionInfo info = {DbErrorOk, "", NULL};

  DbStartup ();
  if (!Initialized) return;
  
  DbClearErrorInfo ();
  
  if ((name == NULL) || (strlen (name) == 0)) 
    {
      DbSetErrorInfo (DbErrorBadArgument, "DbAddTrigger>");
      return;
    }
  
  slash = strrchr (name, '/');
  if (slash == NULL)
    {
      strcpy (object_name, name);
      path_name[0] = 0;
    }
  else
    {
      strcpy (path_name, name);
      slash = strrchr (path_name, '/');
      *slash = 0;
      strcpy (object_name, slash + 1);
    }
  
  if (strlen (object_name) == 0)
    {
      DbSetErrorInfo (DbErrorBadArgument, "DbAddTrigger>");
      return;
    }

  transaction = CmOpenTransaction ("DbAddTrigger", &info);

  message = CmMessageNew ();
  
  CmMessagePutText (message, object_name);
  CmMessagePutText (message, path_name);
  CmMessagePutInt  (message, trigger_type);
  CmMessagePutInt  (message, transaction);

  CmMessageSetType (message, "DbAddTrigger");
  
  if (CmMessageSend (message, DbServerName ()))
    {
      DbWaitForTransaction (transaction);
    }
  else
    {
      DbSetErrorInfo (DbErrorNoServer, "DbAddTrigger>");
    }
  
  CmMessageDelete (message);

  CmCloseTransaction (transaction);
  DbSetErrorInfo (info.error, info.error_text);
}

/*---------------------------------------------------------------------*/
void DbRemoveTrigger (char* name, DbTriggerType trigger_type)
/*---------------------------------------------------------------------*/
{
  char object_name[256];
  char path_name[256];
  char* slash;
  CmMessage  message;
  int transaction;
  TransactionInfo info = {DbErrorOk, "", NULL};

  DbStartup ();
  if (!Initialized) return;
  
  DbClearErrorInfo ();
  
  if ((name == NULL) || (strlen (name) == 0)) 
    {
      DbSetErrorInfo (DbErrorBadArgument, "DbRemoveTrigger>");
      return;
    }
  
  slash = strrchr (name, '/');
  if (slash == NULL)
    {
      strcpy (object_name, name);
      path_name[0] = 0;
    }
  else
    {
      strcpy (path_name, name);
      slash = strrchr (path_name, '/');
      *slash = 0;
      strcpy (object_name, slash + 1);
    }
  
  if (strlen (object_name) == 0)
    {
      DbSetErrorInfo (DbErrorBadArgument, "DbRemoveTrigger>");
      return;
    }

  transaction = CmOpenTransaction ("DbRemoveTrigger", &info);

  message = CmMessageNew ();
  
  CmMessagePutText (message, object_name);
  CmMessagePutText (message, path_name);
  CmMessagePutInt  (message, trigger_type);
  CmMessagePutInt  (message, transaction);

  CmMessageSetType (message, "DbRemoveTrigger");
  
  if (CmMessageSend (message, DbServerName ()))
    {
      DbWaitForTransaction (transaction);
    }
  else
    {
      DbSetErrorInfo (DbErrorNoServer, "DbRemoveTrigger>");
    }
  
  CmMessageDelete (message);

  CmCloseTransaction (transaction);
  DbSetErrorInfo (info.error, info.error_text);
}

/*---------------------------------------------------------------------*/
void DbDeleteObject (char* name)
/*---------------------------------------------------------------------*/
{
  CmMessage  message;
  char       object_name[256];
  char       path_name[256];
  char*      slash;
  int transaction;
  TransactionInfo info = {DbErrorOk, "", NULL};
  
  DbStartup ();
  if (!Initialized) return;
  
  if ((name == NULL) ||
      (strlen (name) == 0)) 
    {
      DbSetErrorInfo (DbErrorBadArgument, "DbDeleteObject>");
      return;
    }
  
  slash = strrchr (name, '/');
  if (slash == NULL)
    {
      strcpy (object_name, name);
      path_name[0] = 0;
    }
  else
    {
      strcpy (path_name, name);
      slash = strrchr (path_name, '/');
      *slash = 0;
      strcpy (object_name, slash + 1);
    }
  
  if (strlen (object_name) == 0)
    {
      DbSetErrorInfo (DbErrorBadArgument, "DbDeleteObject>");
      return;
    }
  
  DbClearErrorInfo ();
  
  transaction = CmOpenTransaction ("DbDeleteObject", &info);

  message = CmMessageNew ();
  
  CmMessagePutText (message, object_name);
  CmMessagePutText (message, path_name);
  CmMessagePutInt  (message, transaction);

  CmMessageSetType (message, "DbDeleteObject");
  
  if (CmMessageSend (message, DbServerName ()))
    {
      DbWaitForTransaction (transaction);
    }
  else
    {
      DbSetErrorInfo (DbErrorNoServer, "DbDeleteObject>");
    }
  
  CmMessageDelete (message);

  CmCloseTransaction (transaction);
  DbSetErrorInfo (info.error, info.error_text);
  
  if (DbGetError () == DbErrorOk)
    {
      DbObject object;
      
      object = DbObjectFind (object_name);
      if (object != NULL)
        {
          DbObjectRemove (object);
          GCCleanup ();
        }
    }
}

/*---------------------------------------------------------------------*/
void DbDeleteType (char* name)
/*---------------------------------------------------------------------*/
{
  CmMessage message;
  int transaction;
  TransactionInfo info = {DbErrorOk, "", NULL};
  
  DbStartup ();
  if (!Initialized) return;
  
  if ((name == NULL) ||
      (strlen (name) == 0)) 
    {
      DbSetErrorInfo (DbErrorBadArgument, "DbDeleteType>");
      return;
    }
  
  DbClearErrorInfo ();
  
  transaction = CmOpenTransaction ("DbDeleteType", &info);

  message = CmMessageNew ();
  
  CmMessagePutText (message, name);
  CmMessagePutInt  (message, transaction);

  CmMessageSetType (message, "DbDeleteType");
  
  if (CmMessageSend (message, DbServerName ()))
    {
      DbWaitForTransaction (transaction);
    }
  else
    {
      DbSetErrorInfo (DbErrorNoServer, "DbDeleteType>");
    }
  
  CmMessageDelete (message);

  CmCloseTransaction (transaction);
  DbSetErrorInfo (info.error, info.error_text);
}

/*---------------------------------------------------------------------*/
char* DbObjectGetNames (char* expression)
/*---------------------------------------------------------------------*/
{
  CmMessage message;
  int transaction;
  TransactionInfo info = {DbErrorOk, "", NULL};
  
  DbStartup ();
  if (!Initialized) return (NULL);
  
  NamesLength = 0;
  if (NamesBuffer != NULL) NamesBuffer[0] = 0;
  
  DbClearErrorInfo ();
  
  transaction = CmOpenTransaction ("DbObjectGetNames", &info);

  message = CmMessageNew ();
  
  CmMessagePutText (message, expression);
  CmMessagePutInt  (message, transaction);

  CmMessageSetType (message, "DbObjectGetNames");
  
  if (CmMessageSend (message, DbServerName ()))
    {
      DbWaitForTransaction (transaction);
    }
  else
    {
      DbSetErrorInfo (DbErrorNoServer, "DbObjectGetNames>");
    }
  
  CmMessageDelete (message);

  CmCloseTransaction (transaction);
  DbSetErrorInfo (info.error, info.error_text);
  
  return (NamesBuffer);
}

/*---------------------------------------------------------------------*/
char* DbObjectGetNamesInPath (char* path)
/*---------------------------------------------------------------------*/
{
  CmMessage message;
  int transaction;
  TransactionInfo info = {DbErrorOk, "", NULL};

  DbStartup ();
  if (!Initialized) return (NULL);
  
  NamesLength = 0;
  if (NamesBuffer != NULL) NamesBuffer[0] = 0;
  
  DbClearErrorInfo ();
  
  transaction = CmOpenTransaction ("DbObjectGetNamesInPath", &info);

  message = CmMessageNew ();
  
  CmMessagePutText (message, path);
  CmMessagePutInt  (message, transaction);

  CmMessageSetType (message, "DbOGetNamesInPath");
  
  if (CmMessageSend (message, DbServerName ()))
    {
      DbWaitForTransaction (transaction);
    }
  else
    {
      DbSetErrorInfo (DbErrorNoServer, "DbObjectGetNamesInPath>");
    }
  
  CmMessageDelete (message);

  CmCloseTransaction (transaction);
  DbSetErrorInfo (info.error, info.error_text);
  
  return (NamesBuffer);
}

/*---------------------------------------------------------------------*/
char* DbSubPathGetNames (char* path)
/*---------------------------------------------------------------------*/
{
  CmMessage message;
  int transaction;
  TransactionInfo info = {DbErrorOk, "", NULL};
  
  DbStartup ();
  if (!Initialized) return (NULL);
  
  NamesLength = 0;
  if (NamesBuffer != NULL) NamesBuffer[0] = 0;
  
  DbClearErrorInfo ();
  
  transaction = CmOpenTransaction ("DbSubPathGetNames", &info);

  message = CmMessageNew ();
  
  CmMessagePutText (message, path);
  CmMessagePutInt  (message, transaction);

  CmMessageSetType (message, "DbSubPathGetNames");
  
  if (CmMessageSend (message, DbServerName ()))
    {
      DbWaitForTransaction (transaction);
    }
  else
    {
      DbSetErrorInfo (DbErrorNoServer, "DbSubPathGetNames>");
    }
  
  CmMessageDelete (message);

  CmCloseTransaction (transaction);
  DbSetErrorInfo (info.error, info.error_text);
  
  return (NamesBuffer);
}

/*---------------------------------------------------------------------*/
char* DbTypeGetNames (char* expression)
/*---------------------------------------------------------------------*/
{
  CmMessage message;
  int transaction;
  TransactionInfo info = {DbErrorOk, "", NULL};
  
  DbStartup ();
  if (!Initialized) return (NULL);
  
  NamesLength = 0;
  if (NamesBuffer != NULL) NamesBuffer[0] = 0;
  
  DbClearErrorInfo ();
  
  transaction = CmOpenTransaction ("DbTypeGetNames", &info);

  message = CmMessageNew ();
  
  CmMessagePutText (message, expression);
  CmMessagePutInt  (message, transaction);

  CmMessageSetType (message, "DbTypeGetNames");
  
  if (CmMessageSend (message, DbServerName ()))
    {
      DbWaitForTransaction (transaction);
    }
  else
    {
      DbSetErrorInfo (DbErrorNoServer, "DbTypeGetNames>");
    }
  
  CmMessageDelete (message);

  CmCloseTransaction (transaction);
  DbSetErrorInfo (info.error, info.error_text);
  
  return (NamesBuffer);
}

/*---------------------------------------------------------------------*/
DbType DbLoadType (char* name)
/*---------------------------------------------------------------------*/
{
  DbType     result = NULL;
  CmMessage  message;
  char       type_name[256];
  
  DbStartup ();
  if (!Initialized) return (NULL);
  
  if ((name == NULL) ||
      (strlen (name) == 0)) 
    {
      DbSetErrorInfo (DbErrorBadArgument, "DbLoadType>");
      return (NULL);
    }
  
  strcpy (type_name, name);
  
  for (;;)
    {
      int transaction;
      TransactionInfo info = {DbErrorOk, "", NULL};

      DbClearErrorInfo ();
      
      transaction = CmOpenTransaction ("DbLoadType", &info);

      message = CmMessageNew ();
      
      CmMessagePutText (message, type_name);
      CmMessagePutInt  (message, transaction);

      CmMessageSetType (message, "DbLoadType");

      if (CmMessageSend (message, DbServerName ()))
        {
          DbWaitForTransaction (transaction);
          result = DbTStructureFind (type_name);
        }
      else
        {
          DbSetErrorInfo (DbErrorNoServer, "DbLoadType>");
        }
      
      CmMessageDelete (message);

      CmCloseTransaction (transaction);
      DbSetErrorInfo (info.error, info.error_text);
      
      if (DbGetError () == DbErrorUnloadedStructure)
        {
          GCReferenceAssign (result, DbTStructureFind (type_name));
          DbTStructureUninstall (result);
          GCReferenceClear (result);

          if (DbLoadType (DbGetErrorInfo ()) == NULL) break;
        }
      else
        {
          break;
        }
    }
  
  return (result);
}

/*---------------------------------------------------------------------*/
void DbStopServer ()
/*---------------------------------------------------------------------*/
{
  CmMessage message;
  int transaction;
  TransactionInfo info = {DbErrorOk, "", NULL};
  
  DbStartup ();
  if (!Initialized) return;
  
  DbClearErrorInfo ();
  
  transaction = CmOpenTransaction ("DbStopServer", &info);

  message = CmMessageNew ();
  
  CmMessagePutText (message, "Stop");
  CmMessagePutInt  (message, transaction);

  CmMessageSetType (message, "DbCommand");

  if (CmMessageSend (message, DbServerName ()))
    {
      DbWaitForTransaction (transaction);
    }
  else
    {
      DbSetErrorInfo (DbErrorNoServer, "DbStopServer>");
    }
  
  CmMessageDelete (message);

  CmCloseTransaction (transaction);
  DbSetErrorInfo (info.error, info.error_text);
}

/*---------------------------------------------------------------------*/
void DbRestartServer ()
/*---------------------------------------------------------------------*/
{
  CmMessage message;
  int transaction;
  TransactionInfo info = {DbErrorOk, "", NULL};
  
  DbStartup ();
  if (!Initialized) return;
  
  DbClearErrorInfo ();
  
  transaction = CmOpenTransaction ("DbRestartServer", &info);

  message = CmMessageNew ();
  
  CmMessagePutText (message, "Restart");
  CmMessagePutInt  (message, transaction);

  CmMessageSetType (message, "DbCommand");

  if (CmMessageSend (message, DbServerName ()))
    {
      DbWaitForTransaction (transaction);
    }
  else
    {
      DbSetErrorInfo (DbErrorNoServer, "DbRestartServer>");
    }
  
  CmMessageDelete (message);

  CmCloseTransaction (transaction);
  DbSetErrorInfo (info.error, info.error_text);
}

/*---------------------------------------------------------------------*/
static DbType DbCreateType (char* definition)
/*---------------------------------------------------------------------*/
{
  DbStartup ();
  
  return (DbParserCreateType (definition));
}

/*---------------------------------------------------------------------*/
char* DbReadFile (char* file_name)
/*---------------------------------------------------------------------*/
{
  FILE* file;
  char* text = NULL;
  int   bytes;
  
  file = fopen (file_name, "r");
  
  if (file == NULL)
    {
      char* env;
      char* path;
      char name[256];
      
      env = getenv ("DBOBJECTSROOT");
      if (env != NULL)
        {
          char* copy_env;
          char* path;
          
          copy_env = (char*) malloc (strlen (env) + 1);
          strcpy (copy_env, env);
          
          path = copy_env;
          while (strlen (path) > 0)
            {
              char* space;
              
              while (*path == ' ') path++;
              
              space = strstr (path, " ");
              if (space != NULL) *space = 0;
              
              strcpy (name, path);
              strcat (name, "/");
              strcat (name, file_name);
              file = fopen (name, "r");
              if (file != NULL) break;
              
              if (space == NULL) break;
              path = space;
              path++;
            }
          
          free (copy_env);
        }
    }
  
  if (file == NULL) return (NULL);
  
  if (!fseek(file, 0, 2))
    {
      bytes = ftell (file);
      
      text = (char*) malloc (bytes + 1);
      
      if (text != NULL)
        {
          fseek (file, 0, 0);
          fread (text, bytes, 1, file);
          text[bytes] = 0;
        }
    }
  else
    {
      text = NULL;
    }
  
  fclose (file);
  
  return (text);
}

/*---------------------------------------------------------------------*/
int DbWriteFile (char* name, char* text)
/*---------------------------------------------------------------------*/
{
  FILE* file;
  int   bytes;
  
  file = fopen (name, "w");
  if (!file) return (0);
  
  if (text == NULL) return (0);
  
  bytes = strlen (text);
  
  fwrite (text, bytes, 1, file);
  
  fclose (file);
  
  return (1);
}

/*- Procedure ---------------------------------------------------------*/
char* DbServerName ()
/*---------------------------------------------------------------------*/
{
  char* env = getenv ("DBSERVERNAME");
  static char server[256] = "DbServerv4r3";

  if (env != NULL) strcpy (server, env);
  
  return (server);
}

/*-----------------------------------------------------------*/
void DbSetDebug ()
/*-----------------------------------------------------------*/
{
  Debug = 1;
}

/*-----------------------------------------------------------*/
void DbSetNoDebug ()
/*-----------------------------------------------------------*/
{
  Debug = 0;
}

/*-----------------------------------------------------------*/
int DbDebug ()
/*-----------------------------------------------------------*/
{
  return (Debug);
}


/*-----------------------------------------------------------*/
void DbSetVersioning ()
/*-----------------------------------------------------------*/
{
  CmMessage message;
  int transaction;
  TransactionInfo info = {DbErrorOk, "", NULL};

  DbStartup ();
  if (!Initialized) return;

  DbClearErrorInfo ();

  transaction = CmOpenTransaction ("DbSetVersioning", &info);

  message = CmMessageNew ();

  CmMessagePutInt  (message, transaction);

  CmMessageSetType (message, "DbSetVersioning");

  if (!CmMessageSend (message, DbServerName ()))
    {
      DbSetErrorInfo (DbErrorNoServer, "DbSetVersioning>");
    }

  CmMessageDelete (message);

  CmCloseTransaction (transaction);
  DbSetErrorInfo (info.error, info.error_text);
}


/*-----------------------------------------------------------*/
void DbSetNoVersioning ()
/*-----------------------------------------------------------*/
{
  CmMessage message;
  int transaction;
  TransactionInfo info = {DbErrorOk, "", NULL};

  DbStartup ();
  if (!Initialized) return;

  DbClearErrorInfo ();

  transaction = CmOpenTransaction ("DbSetNoVersioning", &info);

  message = CmMessageNew ();

  CmMessagePutInt  (message, transaction);

  CmMessageSetType (message, "DbSetNoVersioning");

  if (!CmMessageSend (message, DbServerName ()))
    {
      DbSetErrorInfo (DbErrorNoServer, "DbSetNoVersioning>");
    }
  
  CmMessageDelete (message);

  CmCloseTransaction (transaction);
  DbSetErrorInfo (info.error, info.error_text);
}


/*-----------------------------------------------------------*/
int DbGetVersioning ()
/*-----------------------------------------------------------*/
{
  CmMessage  message;
  int transaction;
  TransactionInfo info = {DbErrorOk, "", NULL};

  DbStartup ();
  if (!Initialized) return (0);

  DbClearErrorInfo ();

  transaction = CmOpenTransaction ("DbGetVersioning", &info);

  message = CmMessageNew ();

  CmMessagePutInt  (message, transaction);

  CmMessageSetType (message, "DbGetVersioning");

  UseVersioning = 0;

  if (CmMessageSend (message, DbServerName ()))
    {
      DbWaitForTransaction (transaction);
    }
  else
    {
      DbSetErrorInfo (DbErrorNoServer, "DbGetVersioning>");
    }
  
  CmMessageDelete (message);

  CmCloseTransaction (transaction);
  DbSetErrorInfo (info.error, info.error_text);

  if (DbDebug ())
    {
      printf ("DbGetVersioning>  UseVersioning : %d\n", UseVersioning);
    }

  return (UseVersioning);
}

/*---------------------------------------------------------------------*/
void DbSetLocks (DbLockMode mode)
/*---------------------------------------------------------------------*/
{
  CmMessage message;
  int transaction;
  TransactionInfo info = {DbErrorOk, "", NULL};
  
  DbStartup ();
  if (!Initialized) return;
  
  DbClearErrorInfo ();
  
  transaction = CmOpenTransaction ("DbSetLocks", &info);

  message = CmMessageNew ();
  
  CmMessagePutInt  (message, mode);
  CmMessagePutInt  (message, transaction);

  CmMessageSetType (message, "DbSetLocks");

  if (CmMessageSend (message, DbServerName ()))
    {
      DbWaitForTransaction (transaction);
    }
  else
    {
      DbSetErrorInfo (DbErrorNoServer, "DbSetLocks>");
    }
  
  CmMessageDelete (message);

  CmCloseTransaction (transaction);
  DbSetErrorInfo (info.error, info.error_text);
}

/*---------------------------------------------------------------------*/
void DbResetLocks ()
/*---------------------------------------------------------------------*/
{
  CmMessage message;
  int transaction;
  TransactionInfo info = {DbErrorOk, "", NULL};
  
  DbStartup ();
  if (!Initialized) return;
  
  DbClearErrorInfo ();
  
  transaction = CmOpenTransaction ("DbResetLocks", &info);

  message = CmMessageNew ();
  
  CmMessagePutInt  (message, transaction);

  CmMessageSetType (message, "DbResetLocks");

  if (CmMessageSend (message, DbServerName ()))
    {
      DbWaitForTransaction (transaction);
    }
  else
    {
      DbSetErrorInfo (DbErrorNoServer, "DbResetLocks>");
    }
  
  CmMessageDelete (message);

  CmCloseTransaction (transaction);
  DbSetErrorInfo (info.error, info.error_text);
}


/*- Procedure ---------------------------------------------------------*/
/*                                                                     */
/* The server has just completed its transaction. The message may      */
/* contain an error message in case of troubles during the transaction */
/*                                                                     */
/*---------------------------------------------------------------------*/
static CmMessageStatus BreakHandler (CmMessage message, 
                                     char* sender, char* server)
/*---------------------------------------------------------------------*/
{
  char* text = "";
  DbError error;
  
  if (DbDebug ())
    {
      DbPrint ("BreakHandler> from %s (pending %d)\n", sender,
               TransactionPending);
    }
  
  error = (DbError) CmMessageGetInt (message);
  
  if (error == DbErrorOk)
    {
      text = DbGetErrorInfo ();
      error = DbGetError ();
    }
  else
    {
      text = CmMessageGetText (message);
      DbSetErrorInfo (error, "%s", text);
    }

    /*
      This is a dummy value for backward compatibility
      */

  CmMessageGetInt (message);

  TransactionPending--;

  if (CmMessageGetItemType (message) == CmMessageItemInt)
    {
      int transaction;
      TransactionInfo* info;

      transaction = CmMessageGetInt (message);
      CmTerminateTransaction (transaction);

      info = (TransactionInfo*) CmGetTransactionObject (transaction);
      info->error = error;
      strcpy (info->error_text, text);

      GCCleanup ();

      return (CmMessageOk);
    }
  else
    {
        /*
          This should become obsolete when transaction
          mechanisms are generalised
          This is maintained for backward compatibility
          */
  
      GCCleanup ();

      if (TransactionPending == 0) return (CmMessageBreak);
      else return (CmMessageOk);

      return (CmMessageBreak);
    }
}

/*---------------------------------------------------------------------*/
static CmMessageStatus VersioningHandler (CmMessage message, 
                                          char* sender, char* server)
/*---------------------------------------------------------------------*/
{
  if (DbDebug ())
    {
      DbPrint ("VersioningHandler> from  %s\n", sender);
    }
  
  UseVersioning = CmMessageGetInt (message);
  
  if (DbDebug ())
    {
      DbPrint ("VersioningHandler> versioning : %d\n", UseVersioning);
    }
  
  return (CmMessageOk);
}

/*---------------------------------------------------------------------*/
static CmMessageStatus TypeDefinitionHandler (CmMessage message, 
                                              char* sender, char* server)
/*---------------------------------------------------------------------*/
{
  char* text;
  DbType type;
  
  text = CmMessageGetText (message);
  
  type = DbParserCreateType (text);
  DbTStructureInstall (type);
  
  return (CmMessageOk);
}

/*---------------------------------------------------------------------*/
static CmMessageStatus TypeHeadersHandler (CmMessage message, 
                                           char* sender, char* server)
/*---------------------------------------------------------------------*/
{
  while (CmMessageGetItemType (message) == CmMessageItemText)
    {
      char* name;
      char* text;
      
      name = CmMessageGetText (message);
      text = CmMessageGetText (message);
      
      if (text != NULL)
        {
          DbWriteFile (name, text);
        }
    }
  
  return (CmMessageOk);
}

/*---------------------------------------------------------------------*/
static CmMessageStatus ObjectDefinitionHandler (CmMessage message, 
                                                char* sender, char* server)
/*---------------------------------------------------------------------*/
{
  char* text;
  DbObject object;
  
  text = CmMessageGetText (message);
  
  object = DbParserCreateObject (text);
  DbObjectInstall (object);
  
  return (CmMessageOk);
}

/*---------------------------------------------------------------------*/
static CmMessageStatus ObjectDefsHandler (CmMessage message, 
                                          char* sender, char* server)
/*---------------------------------------------------------------------*/
{
  CLList list = NULL;
  int transaction = 0;

  if (CmMessageGetItemType (message) == CmMessageItemInt)
    {
      TransactionInfo* info;

      transaction = CmMessageGetInt (message);

      info = (TransactionInfo*) CmGetTransactionObject (transaction);
      list = (CLList) info->object;
    }

  if (list != NULL)
    {
      CLListErase (list);
    }

  while (CmMessageGetItemType (message) == CmMessageItemText)
    {
      char* text;
      
      text = CmMessageGetText (message);
      
      if (text != NULL)
        {
          DbObject object = NULL;
          
          if (DbDebug ())
            {
              printf ("ObjectDefsHandler> object : %s\n", text);
            }
          
          object = DbParserCreateObject (text);
          
          if (object != NULL)
            {
                /* Installation des objets en memoire */
              DbObjectInstall (object);
                /* et dans le buffer de retour */
              if (list != NULL) CLListAdd (list, object);
            }
        }
    }
  
  return (CmMessageOk);
}

/*---------------------------------------------------------------------*/
static CmMessageStatus ObjectNamesHandler (CmMessage message, 
                                           char* sender, char* server)
/*---------------------------------------------------------------------*/
{
  char* text;
  int length;
  
  while (CmMessageGetItemType (message) == CmMessageItemText)
    {
      text = CmMessageGetText (message);
      
      if (text != NULL)
        {
          length = strlen (text);
          
          while ((NamesLength + length + 2) > NamesAllocated)
            {
              NamesAllocated += 1000;
              
              if (NamesBuffer == NULL)
                {
                  NamesBuffer = (char*) malloc (NamesAllocated);
                  NamesBuffer[0] = 0;
                }
              else
                {
                  NamesBuffer = (char*) realloc (NamesBuffer, NamesAllocated);
                }
            }
          strcat (NamesBuffer, text);
          strcat (NamesBuffer, " ");
          NamesLength += length + 1;
        }
    }
  
  return (CmMessageOk);
}

/*---------------------------------------------------------------------*/
static CmMessageStatus TypeNamesHandler (CmMessage message, 
                                         char* sender, char* server)
/*---------------------------------------------------------------------*/
{
  char* text;
  int length;
  
  while (CmMessageGetItemType (message) == CmMessageItemText)
    {
      text = CmMessageGetText (message);
      
      if (text != NULL)
        {
          length = strlen (text);
          
          while ((NamesLength + length + 2) > NamesAllocated)
            {
              NamesAllocated += 1000;
              
              if (NamesBuffer == NULL)
                {
                  NamesBuffer = (char*) malloc (NamesAllocated);
                  NamesBuffer[0] = 0;
                }
              else
                {
                  NamesBuffer = (char*) realloc (NamesBuffer, NamesAllocated);
                }
            }
          strcat (NamesBuffer, text);
          strcat (NamesBuffer, " ");
          NamesLength += length + 1;
        }
    }
  
  return (CmMessageOk);
}


/*---------------------------------------------------------------------*/
static CmMessageStatus VersionHistoryHandler (CmMessage message, 
                                              char* sender, char* server)
/*---------------------------------------------------------------------*/
{
  char* text;
  
  while (CmMessageGetItemType (message) == CmMessageItemText)
    {
      text = CmMessageGetText (message);
      
      if ((text != NULL) && (strlen (text) > 0))
        {
          int index;
          DbString* string;

          if (DbDebug ())
            {
              printf ("VersionHistoryHandler> text : %s\n", text);
            }

          if (HistoryBuffer == NULL)
            {
              HistoryBuffer = DbStringVectorNew ();
            }

          index = DbVectorGetSize (HistoryBuffer);

          string = DbStringVectorGetElement (HistoryBuffer, index);
          DbStringSetContents (string, text);
        }
    }
  
  return (CmMessageOk);
}


/*---------------------------------------------------------------------*/
static CmMessageStatus ObjectHistoryHandler (CmMessage message, 
                                             char* sender, char* server)
/*---------------------------------------------------------------------*/
{ 
  return (CmMessageOk);
}

static int* TransactionStack = NULL;
static int  TransactionStackSize = 0;
static int  TransactionsAvailable = 0;

/*---------------------------------------------------------------------*/
static void add_transaction (int transaction)
/*---------------------------------------------------------------------*/
{
  int i;

  if (TransactionsAvailable < 1)
    {
      TransactionStackSize += 100;
      TransactionsAvailable += 100;

      if (TransactionStack == NULL)
        {
          TransactionStack = (int*) malloc (TransactionStackSize * sizeof (int));
        }
      else
        {
          TransactionStack = (int*) realloc (TransactionStack,
                                             TransactionStackSize * sizeof (int));
        }

      for (i = TransactionStackSize - 100; i < TransactionStackSize; i++)
        {
          TransactionStack[i] = 0;
        }
    }

  for (i = 0; i < TransactionStackSize; i++)
    {
      if (TransactionStack[i] == 0)
        {
          TransactionStack[i] = transaction;
          TransactionsAvailable--;
          break;
        }
    }
}

/*---------------------------------------------------------------------*/
static void remove_transaction (int transaction)
/*---------------------------------------------------------------------*/
{
  int i;

  for (i = 0; i < TransactionStackSize; i++)
    {
      if (TransactionStack[i] == transaction)
        {
          TransactionStack[i] = 0;
          TransactionsAvailable++;
          break;
        }
    }
}

/*---------------------------------------------------------------------*/
static void terminate_transactions ()
/*---------------------------------------------------------------------*/
{
  int i;

  for (i = 0; i < TransactionStackSize; i++)
    {
      if (TransactionStack[i] != 0)
        {
          CmTerminateTransaction (TransactionStack[i]);
          TransactionStack[i] = 0;
          TransactionsAvailable++;
        }
    }
}

/*---------------------------------------------------------------------*/
void DbWaitForTransaction (int transaction)
/*---------------------------------------------------------------------*/
{
  int   timeout     = 0;
  char* timeout_str = NULL;
  
  CmConnectCondition condition;
  
  timeout_str = getenv ("DBTIMEOUT");
  if (timeout_str != NULL)
    {
      sscanf (timeout_str, "%d", &timeout);
    }

  TransactionPending++;

  add_transaction (transaction);

  while (!CmIsTransactionTerminated (transaction))
    {
      if (DbDebug ())
        {
          DbPrint ("DbWaitLoop(1)> t=%d (%d pending)\n", transaction, TransactionPending);
          CmDumpTransactions ();
        }
      
      if (timeout == 0)
        condition = CmConnectWait ();
      else
        condition = CmConnectWaitWithTimeout ((double) timeout);
      
      switch (condition)
        {
          case CmConnectNoHandler :
            DbSetErrorInfo (DbErrorNoHandler, "Unknown message type.");
            CmTerminateTransaction (transaction);
            break;
          case CmConnectTimeoutDetection :
            DbSetErrorInfo (DbErrorConnectionLost, 
                            "Connection to Server lost.");
            CmTerminateTransaction (transaction);
            break;
          case CmConnectBreakDetection :
            break;
          case CmConnectErrorCondition :
            DbSetErrorInfo (DbErrorCmError, "Cm error.");
            CmTerminateTransaction (transaction);
            break;
        }

      if (DbDebug ())
        {
          DbPrint ("DbWaitLoop(2)> t=%d (%d pending)\n", transaction, TransactionPending);
          CmDumpTransactions ();
        }

      /*if (TransactionPending == 0) break;*/
    }

  remove_transaction (transaction);
}


/*------------------------------------------------------*/
static void DisconnectionHandler (CmConnect connect)
/*------------------------------------------------------*/
{
  char* name;

  name = CmConnectGetName (connect);

  if ((name != NULL) && !strcmp (name, DbServerName ()))
    {
      terminate_transactions ();
    }

  if (PreviousDisconnectionHandler != NULL)
    PreviousDisconnectionHandler (connect);

}

