/*---------------------------------------------------------------------*\
| Author : ARNAULT Christian                                            |
|          MASSART Pierre                                               |
|                                                                       |
| arnault@lalcls.in2p3.fr                                               |
| Orsay, Laboratoire d'Accelerateur Lineaire (L.A.L.)                   |
|                                                                       |
|                                                                       |
|  DDD     AA   TTTTTTTTT   AA        BBB        AA     SSSSS  EEEEEE   |
|  D  D   A  A      T      A  A       B  B      A  A   S     S E     E  |
|  D   D  A  A      T      A  A       B   B     A  A   S       E        |
|  D   D A    A     T     A    A  __  B BB     A    A   SSSSS  EEEE     |
|  D   D AAAAAA     T     AAAAAA      B   B    AAAAAA        S E        |
|  D  D A      A    T    A      A     B    B  A      A S     S E     E  |
|  DDD  A      A    T    A      A     BBBBB   A      A  SSSSS  EEEEEE   |
|                                                                       |
|                                                                       |
\*---------------------------------------------------------------------*/

/*- Header ------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <Reg.h>
#include <DbArray.h>
/*---------------------------------------------------------------------*/

/*----------------------------------------------------------*/
/*          Array                                           */
/*----------------------------------------------------------*/

typedef void   (*DbIteratorMethod) (DbIterator This, ...);

typedef struct _DbArrayRec
{
  void** array;
  int allocated;
  int used;
  int free;
} DbArrayRec;

#define QUANTUM 20

/*----------------------------------------------------------*/
DbArray DbArrayNew ()
/*----------------------------------------------------------*/
{
  DbArray This;

  This = (DbArray) malloc (sizeof (DbArrayRec));

  DbArrayInit (This);

  return (This);
}

/*----------------------------------------------------------*/
void DbArrayDelete (DbArray This)
/*----------------------------------------------------------*/
{
  if (!This) return;

  DbArrayCleanup (This);

  free (This);
}

/*----------------------------------------------------------*/
void DbArrayInit (DbArray This)
/*----------------------------------------------------------*/
{
  if (!This) return;

  This->array     = NULL;
  This->allocated = 0;
  This->used      = 0;
  This->free      = 0;
}

/*----------------------------------------------------------*/
void DbArrayCleanup (DbArray This)
/*----------------------------------------------------------*/
{
  if (!This) return;

  if (This->array)
    {
      int pos;

      for (pos = 0; pos < This->used; pos++)
	{
	  This->array[pos] = NULL;
	}
      This->used      = 0;
      This->free      = This->allocated;

      free (This->array);
      This->array     = NULL;
      This->allocated = 0;
      This->used      = 0;
      This->free      = 0;
    }
}

/*----------------------------------------------------------*/
int DbArrayCount (DbArray This)
/*----------------------------------------------------------*/
{
  int pos;
  int count;

  if (!This) return (0);
  if (!This->array) return (0);

  count = 0;
  for (pos = 0; pos < This->used; pos++)
    {
      if (This->array[pos] != NULL) count++;
    }

  return (count);
}

/*----------------------------------------------------------*/
void DbArrayAdd (DbArray This, void* object)
/*----------------------------------------------------------*/
{
  int pos;

  if (!This) return;

  if (This->free > 0)
    {
      for (pos = 0; pos < This->used; pos++)
	{
	  if (This->array[pos] == NULL)
	    {
	      This->free--;
	      break;
	    }
	}
    }
  else
    {
      if (This->used >= This->allocated)
	{
	  This->allocated += QUANTUM;
	  if (!This->array)
	    {
	      This->array =
		(void**) malloc (This->allocated *
				    sizeof (void*));
	    }
	  else
	    {
	      This->array =
		(void**) realloc (This->array,
				     This->allocated *
				     sizeof (void*));
	    }
	}
      pos = This->used;
      This->used++;
    }
  This->array[pos] = object;
}

/*----------------------------------------------------------*/
void DbArrayRemove (DbArray This, void* object)
/*----------------------------------------------------------*/
{
  int pos;

  if (!This) return;
  if (!This->array) return;

  for (pos = 0; pos < This->used; pos++)
    {
      if (This->array[pos] == object)
	{
	  This->array[pos] = NULL;
	  This->free++;
	  break;
	}
    }
}

/*----------------------------------------------------------*/
void* DbArrayItem (DbArray This, DbIterator it)
/*----------------------------------------------------------*/
{
  int index;

  if (!This) return (NULL);
  if (!This->array) return (NULL);
  if (!it) return (NULL);
  if (DbIteratorFinished(it)) return (NULL);

  index = DbIteratorIndex (it);

  if (index < 0) return (NULL);
  if (index >= This->used) return (NULL);

  return (This->array[index]);
}

/*----------------------------------------------------------*/
/*    Iterator                                              */
/*----------------------------------------------------------*/

typedef struct _DbIteratorClassRec* DbIteratorClass;

typedef struct _DbIteratorClassRec
{
  char*            name;
  DbIteratorMethod forth;
} DbIteratorClassRec;

typedef struct _DbIteratorRec
{
  DbIteratorClass Class;

  DbArray array;
  int current;
  int finished;
} DbIteratorRec;

static void DbForwardIteratorForth (DbIterator This);

static DbIteratorClassRec DbForwardIteratorClass = 
{
  "forward",
  (DbIteratorMethod) DbForwardIteratorForth
};

static void DbBackwardIteratorForth (DbIterator This);

static DbIteratorClassRec DbBackwardIteratorClass = 
{
  "backward",
  (DbIteratorMethod) DbBackwardIteratorForth
};

/*----------------------------------------------------------*/
DbIterator DbForwardIteratorNew (DbArray array)
/*----------------------------------------------------------*/
{
  DbIterator This;

  This = (DbIterator) malloc (sizeof (DbIteratorRec));

  This->Class = &DbForwardIteratorClass;
  This->array = array;
  DbIteratorFirst (This);

  return (This);
}

/*----------------------------------------------------------*/
DbIterator DbBackwardIteratorNew (DbArray array)
/*----------------------------------------------------------*/
{
  DbIterator This;

  This = (DbIterator) malloc (sizeof (DbIteratorRec));

  This->Class = &DbBackwardIteratorClass;
  This->array = array;
  DbIteratorFirst (This);

  return (This);
}

/*----------------------------------------------------------*/
void DbIteratorDelete (DbIterator This)
/*----------------------------------------------------------*/
{
  if (!This) return;

  This->array = NULL;
  This->finished = 1;
  This->current = -1;

  free (This);
}

/*----------------------------------------------------------*/
void DbIteratorInitForward (DbIterator This, DbArray array)
/*----------------------------------------------------------*/
{
  if (!This) return;

  This->Class = &DbForwardIteratorClass;
  This->array = array;
  DbIteratorFirst (This);
}

/*----------------------------------------------------------*/
void DbIteratorInitBackward (DbIterator This, DbArray array)
/*----------------------------------------------------------*/
{
  if (!This) return;

  This->Class = &DbBackwardIteratorClass;
  This->array = array;
  DbIteratorFirst (This);
}

/*----------------------------------------------------------*/
int DbIteratorFinished (DbIterator This)
/*----------------------------------------------------------*/
{
  if (!This) return (1);

  return (This->finished);
}

/*----------------------------------------------------------*/
int DbIteratorIndex (DbIterator This)
/*----------------------------------------------------------*/
{
  if (!This) return (-1);

  return (This->current);
}

/*----------------------------------------------------------*/
void DbIteratorFirst (DbIterator This)
/*----------------------------------------------------------*/
{
  if (!This) return;

  This->finished = 0;
  This->current  = -1;

  DbIteratorForth (This);
}

/*----------------------------------------------------------*/
void DbIteratorForth (DbIterator This)
/*----------------------------------------------------------*/
{
  if (!This) return;
  if (This->finished) return;

  if (This->Class->forth) This->Class->forth (This);
}

/*----------------------------------------------------------*/
static void DbForwardIteratorForth (DbIterator This)
/*----------------------------------------------------------*/
{
  This->finished = 1;

  if (This->array)
    {
      if (This->array->array)
	{
	  int pos;

	  This->current++;

	  for (pos = This->current; pos < This->array->used; pos++)
	    {
	      if (This->array->array[pos] != NULL)
		{
		  This->finished = 0;
		  This->current  = pos;
		  break;
		}
	    }
	}
    }
}

/*----------------------------------------------------------*/
static void DbBackwardIteratorForth (DbIterator This)
/*----------------------------------------------------------*/
{
  This->finished = 1;

  if (This->array)
    {
      if (This->array->array)
	{
	  int pos;

	  if (This->current < 0) This->current = This->array->used;
	  This->current--;

	  for (pos = This->current; pos >= 0; pos--)
	    {
	      if (This->array->array[pos] != NULL)
		{
		  This->finished = 0;
		  This->current  = pos;
		  break;
		}
	    }
	}
    }
}

