/*---------------------------------------------------------------------------*/
/*        Copyright (c) 1996 LAL Orsay, UPS-IN2P3-CNRS (France).             */
/*                                                                           */
/* Redistribution and use in source and binary forms, with or without        */
/* modification, are permitted provided that the following conditions        */
/* are met:                                                                  */
/* 1. Redistributions of source code must retain the above copyright         */
/*    notice, this list of conditions and the following disclaimer.          */
/* 2. Redistributions in binary form must reproduce the above copyright      */
/*    notice, this list of conditions and the following disclaimer in the    */
/*    documentation and/or other materials provided with the distribution.   */
/* 3. All advertising materials mentioning features or use of this software  */
/*    must display the following acknowledgement:                            */
/*      This product includes software developed by the Computer Application */
/*      Development Group at LAL Orsay (Laboratoire de l'Accelerateur        */
/*      Linaire - UPS-IN2P3-CNRS).                                           */
/* 4. Neither the name of the Institute nor of the Laboratory may be used    */
/*    to endorse or promote products derived from this software without      */
/*    specific prior written permission.                                     */
/*                                                                           */
/* THIS SOFTWARE IS PROVIDED BY THE LAL AND CONTRIBUTORS ``AS IS'' AND       */
/* ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE     */
/* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR        */
/* PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE LAL OR CONTRIBUTORS BE      */
/* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR       */
/* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF      */
/* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS  */
/* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN   */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)   */
/* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF    */
/* THE POSSIBILITY OF SUCH DAMAGE.                                           */
/*---------------------------------------------------------------------------*/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <CSet.h>
#include <DbDefs.h>
#include <DbObjectPath.h>

/*---------------------------------------------------------------------------*/
typedef struct _DbObjectPathRec
{
  char* name;
  CHasher objects;
} DbObjectPathRec;
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
static void DbObjectPathDelete (DbObjectPath path);
static void DbObjectPathErase (DbObjectPath path);
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
static CHasher Paths = NULL;
static GCClass DbObjectPathGCClass = NULL;
static int     Initialized = 0;
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
static void DbObjectPathStartup ()
/*---------------------------------------------------------------------------*/
{
  if (!Initialized)
    {
      Initialized = 1;
      DbObjectPathNew ("");
    }
}

/*---------------------------------------------------------------------------*/
void DbObjectPathCleanup ()
/*---------------------------------------------------------------------------*/
{
  DbObjectPath path;
  CIterator it;

  it = CHasherGiveIterator (Paths);
  for (; !CIteratorFinished (it); CIteratorForth (it))
    {
      path = (DbObjectPath) CIteratorItem (it);
      DbObjectPathErase (path);

      CHasherRemove (Paths, path);
    }

  CIteratorDelete (it);

  GCReferenceClear (Paths);
  GCReferenceClear (DbObjectPathGCClass);

  GCCleanup ();
}

/*---------------------------------------------------------------------------*/
DbObjectPath DbObjectPathFind (char* name)
/*---------------------------------------------------------------------------*/
{
  DbObjectPath path;

  if (name == NULL) name = "";

  DbObjectPathStartup ();

  if (Paths == NULL) return (NULL);

  path = (DbObjectPath) CHasherFind (Paths, name);

  return (path);
}

/*---------------------------------------------------------------------------*/
DbObjectPath DbObjectPathNew (char* name)
/*---------------------------------------------------------------------------*/
{
  DbObjectPath path;

  if (name == NULL) name = "";

  DbObjectPathStartup ();

  path = DbObjectPathFind (name);
  if (path != NULL)
    {
      return (path);
    }

    /*
      The path is not yet defined
      */

  path = (DbObjectPath) malloc (sizeof (DbObjectPathRec));
  path->name = (char*) malloc (strlen (name) + 1);
  strcpy (path->name, name);
  path->objects = NULL;

  if (DbObjectPathGCClass == NULL)
    {
      GCReferenceAssign (DbObjectPathGCClass,
                         GCClassNew ("DbObjectPath", 
                                     (GCDeleteMethod) DbObjectPathDelete));
    }

  GCObjectNew (path, DbObjectPathGCClass);

  if (Paths == NULL)
    {
      GCReferenceAssign (Paths, 
                         CStrHasherNew (100, (CValuator) DbObjectPathGetName));
    }

  CHasherAdd (Paths, path);

  return (path);
}

/*---------------------------------------------------------------------------*/
void DbObjectPathRemove (DbObjectPath path)
/*---------------------------------------------------------------------------*/
{
  if (path == NULL) return;

  CHasherRemove (Paths, path);

  GCCleanup ();
}

/*---------------------------------------------------------------------------*/
static void DbObjectPathErase (DbObjectPath path)
/*---------------------------------------------------------------------------*/
{
  if (path == NULL) return;

  if (path->objects != NULL)
    {
      CHasherErase (path->objects);

      GCReferenceClear (path->objects);
    }
}

/*---------------------------------------------------------------------------*/
static void DbObjectPathDelete (DbObjectPath path)
/*---------------------------------------------------------------------------*/
{
  if (path == NULL) return;

  DbObjectPathErase (path);

  if (path->name != NULL)
    {
      free (path->name);
      path->name = NULL;
    }

  free (path);
}

/*---------------------------------------------------------------------------*/
char* DbObjectPathGetName (DbObjectPath path)
/*---------------------------------------------------------------------------*/
{
  if (path == NULL) return (NULL);

  return (path->name);
}

/*---------------------------------------------------------------------------*/
CIterator DbObjectPathGiveObjectIterator (DbObjectPath path)
/*---------------------------------------------------------------------------*/
{
  if (path == NULL) return (NULL);
  if (path->objects == NULL) return (NULL);

  return (CHasherGiveIterator (path->objects));
}

/*---------------------------------------------------------------------------*/
DbObject DbObjectPathFindObject (DbObjectPath path, char* name)
/*---------------------------------------------------------------------------*/
{
  DbObject object;

  if (path == NULL) return (NULL);
  if (path->objects == NULL) return (NULL);

  if (name == NULL) return (NULL);

  object = (DbObject) CHasherFind (path->objects, name);

  return (object);
}

/*---------------------------------------------------------------------------*/
int DbObjectPathInstallObject (DbObjectPath path, DbObject object)
/*---------------------------------------------------------------------------*/
/*                                                                     */
/*   Returns (1) if a true installation occured                        */
/*  or (0) if the object was already installed.                        */
/*                                                                     */
/*---------------------------------------------------------------------*/
{
  if (path == NULL) return (0);
  if (object == NULL) return (0);

  if (DbObjectPathFindObject (path, DbObjectGetName (object)) != NULL)
    {
      return (0);
    }

  if (path->objects == NULL)
    {
      GCReferenceAssignIn (path->objects, path,
                           CStrHasherNew (100, (CValuator) DbObjectGetName));
    }

  CHasherAdd (path->objects, object);

  return (1);
}

/*---------------------------------------------------------------------------*/
int DbObjectPathRemoveObject (DbObjectPath path, DbObject object)
/*---------------------------------------------------------------------------*/
/*                                                                     */
/*   Returns (1) if a true removal occured                             */
/*  or (0) otherwise                          .                        */
/*                                                                     */
/*---------------------------------------------------------------------*/
{
  if (path == NULL) return (0);
  if (object == NULL) return (0);
  if (path->objects == NULL) return (0);

  CHasherRemove (path->objects, object);

  return (1);
}

