%{

/*-- Include files ----------------------------------------------*/
#include <stdarg.h>
#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#define alloca malloc
/*---------------------------------------------------------------*/

/*---------------------------------------------------------------*/
#include <Db.h>
/*---------------------------------------------------------------*/


/*-- Prototypes -------------------------------------------------*/
#define yyparse DbSemanticAnalyzer
#define yyerror DbLexicalError
#define yyerrlab DbErrLbl
#define yyerrflag DbErrFlg

  int yylex();

int DbSemanticAnalyzer ();
int DbYYlex ();

typedef enum
{
  ObjectHead,
  ObjectCopy
} ParserMode;

typedef struct
{
  DbVersioningOption option;
  ParserMode       mode;
  DbObject         first_object;
  DbObject         object;
  DbType           type;
  DbTypeIterator   type_iterator;
  CLList           objects;  /* For multiple object definition */
  char*            name;
} ParserContext;

typedef struct
{
  char* buffer;
  int length;
  int current_position;
  int line_number;
  char* old_line_start;
  char* line_start;
} LexContext;

static ParserContext* Context = NULL;
static LexContext*    Lex_context = NULL;
 
int DbLexLineNumber ();
char* DbLexLineStart ();
int DbLexPos ();
void DbLexRestoreContext (LexContext* context);
LexContext* DbLexStartup (LexContext* context, char* text);

static int DbLexicalError(char* s);

static void ContextSetName (ParserContext* context, char* name);
static void ContextAppendName (ParserContext* context, char* name);
/*---------------------------------------------------------------*/
%}

%union
{
  DbVersioningOption option;
  CLList      list;
  DbObject    object;
  DbType      type;
  double      real;
  int         number;
  char*       string;
}

%start db

%token OPEN_SQUARE
%token CLOSE_SQUARE
%token OPEN_BRACE
%token CLOSE_BRACE
%token OPEN_PAR
%token CLOSE_PAR
%token LOWER
%token GREATER
%token COMMA
%token EQUAL
%token SLASH

%token SHORT
%token INT
%token HEXA
%token FLOAT
%token DOUBLE

%token CHARACTER
%token STRING

%token REFERENCE

%token VERSION
%token NOVERSION
%token AT

%token <string> NAME
%token <string> TEXT
%token <real> REAL
%token <number> NUMBER

%type <type> object_type
%type <type> basic_type
%type <type> set
%type <object> object
%type <object> object_declaration

%type <string> object_name
%type <string> opt_owner

%type <option> opt_versioning

%%

db :
  /* empty */
  | objects
    {
    }
  ;

objects :
    object
    {
      DbObject object = $1;
      
      if ((Context->objects != NULL) && (object != NULL))
      {
        CLListAdd (Context->objects, object);
      }
      if (Context->first_object == NULL) Context->first_object = object;
    }
  | objects object
    {
      DbObject object = $2;
      
      if ((Context->objects != NULL) && (object != NULL))
      {
        CLListAdd (Context->objects, object);
      }
      if (Context->first_object == NULL) Context->first_object = object;
    }
  ;

object_name : NAME
    {
      char* name = $1;

      ContextSetName (Context, name);

      $$ = Context->name;
    }
  | object_name SLASH NAME
    {
      char* path = $1;
      char* name = $3;

      ContextAppendName (Context, name);

      $$ = Context->name;
    }
  ;

object :
    object_declaration
    {
      DbObject object = $1;

      $$ = object;
    }
  | object_declaration EQUAL
    {
      DbObject object = $1;

      DbObjectClear (object);

      Context->type_iterator = DbObjectGiveOutputIterator (object);
    }
    opt_value
    {
      DbTypeIteratorDelete (Context->type_iterator);
      Context->type_iterator = NULL;

      $$ = Context->object;
    }
  ;

opt_versioning :
  /* empty */
    {
      $$ = DbDefaultVersioning;
    }
  | OPEN_PAR VERSION CLOSE_PAR
    {
      $$ = DbUseVersioning;
    }
  | OPEN_PAR NOVERSION CLOSE_PAR
    {
      $$ = DbNoVersioning;
    }
  ;

opt_owner :
  /* empty */
    {
      $$ = NULL;
    }
  | AT NAME
    {
      char* name = $2;

      $$ = name;
    }
  ;

object_declaration :
    opt_versioning object_type object_name opt_owner
    {
      DbVersioningOption versioning = $1;
      DbType type = $2;
      char* name = $3;
      char* owner = $4;

      DbObject object = NULL;
      char* path;
      char* slash;

      slash = strrchr (name, '/');
      if (slash == NULL) path = NULL;
      else
        {
          *slash = 0;
          path = name;
          name = slash + 1;
        }

      if ((Context->mode == ObjectHead) &&
          (type != NULL)) 
        {
          object = DbObjectFindInPath (name, path);
          if (object == NULL)
            {
              object = DbObjectNewInPath (type, name, path);
            }
        }

      if ((Context->mode == ObjectCopy) &&
          (type != NULL)) 
        {
          object = DbObjectNewInPath (type, name, path);
        }

      DbObjectSetVersioningOption (object, versioning);
      DbObjectSetOwner (object, owner);

      Context->object = object;
      
      $$ = object;
    }
  ;

object_type :
    basic_type
    {
      $$ = $1;
    }
  | set
    {
      $$ = $1;
    }
  | error
    {
      DbLexicalError ("Bad object type specification..");

      $$ = NULL;
    }
  ;

basic_type :
    CHARACTER
    {
      $$ = DbTCharNew ();
    }
  | SHORT
    {
      $$ = DbTShortNew ();
    }
  | INT
    {
      $$ = DbTIntNew ();
    }
  | HEXA
    {
      $$ = DbTHexaNew ();
    }
  | FLOAT
    {
      $$ = DbTFloatNew ();
    }
  | DOUBLE
    {
      $$ = DbTDoubleNew ();
    }
  | STRING
    {
      $$ = DbTStringNew ();
    }
  | REFERENCE
    {
      $$ = DbTReferenceNew (NULL);
    }
  | REFERENCE LOWER basic_type GREATER
    {
      DbType type = $3;

      $$ = DbTReferenceNew (type);
    }
  | NAME
    {
      char* name = $1;
      DbType type;

      type = DbTStructureFind (name);

      if (type == NULL)
        {
            /*
              Unknown user type
              */
          DbSetErrorInfo (DbErrorUnloadedStructure, "%s", name);
        }

      $$ = type;
    }
  ;

set :
    basic_type OPEN_SQUARE CLOSE_SQUARE
    {
      DbType type = $1;

      $$ = DbTVectorNew (type);
    }
  | basic_type OPEN_SQUARE NUMBER CLOSE_SQUARE OPEN_SQUARE NUMBER CLOSE_SQUARE
    {
      DbType type = $1;
      int rows = $3;
      int columns = $6;

      $$ = DbTMatrixNew (type, rows, columns);
    }
  | basic_type OPEN_SQUARE REAL CLOSE_SQUARE OPEN_SQUARE REAL CLOSE_SQUARE
    {
      DbType type = $1;
      int rows = (int) $3;
      int columns = (int) $6;

      $$ = DbTMatrixNew (type, rows, columns);
    }
  | basic_type OPEN_SQUARE error CLOSE_SQUARE
    {
      DbType type = $1;

      DbLexicalError ("Bad matrix specification");

      $$ = NULL;
    }
  ;

opt_value :
  /* empty */
  | value
  ;

value :
    direct_value
  | keyed_value
  ;

direct_value :
    NUMBER
    {
      int n = $1;

      DbTypeIteratorStoreInt (Context->type_iterator, n);
      DbTypeIteratorForth (Context->type_iterator);
    }
  | REAL
    {
      double d = $1;

      DbTypeIteratorStoreDouble (Context->type_iterator, d);
      DbTypeIteratorForth (Context->type_iterator);
    }
  | TEXT
    {
      char* text = $1;

      DbTypeIteratorStoreText (Context->type_iterator, text);
      DbTypeIteratorForth (Context->type_iterator);
    }
  | object_name
    {
      char* text = $1;

      DbTypeIteratorStoreName (Context->type_iterator, text);
      DbTypeIteratorForth (Context->type_iterator);
    }
  | OPEN_BRACE 
    {
      DbTypeIterator sub;

      sub = DbTypeIteratorDown (Context->type_iterator);

      Context->type_iterator = sub;
    }
    opt_values CLOSE_BRACE
    {
      Context->type_iterator = DbTypeIteratorUp (Context->type_iterator);
      DbTypeIteratorForth (Context->type_iterator);
    }
  | OPEN_BRACE error CLOSE_BRACE
    {
      DbLexicalError ("Bad field value specification");
    }
  ;

keyed_value :
    NAME EQUAL
    {
      char* name = $1;

      DbTypeIteratorMoveTo (Context->type_iterator, name);
    }
    direct_value
  ;

opt_values :
  /* empty */
  | values
  ;

values :
    value
  | values value
  ;

%%

/*---------------------------------------------------------------*/
int yywrap ()
/*---------------------------------------------------------------*/
{
  return (1);
}

/*---------------------------------------------------------------*/
static int DbLexicalError (char* s)
/*---------------------------------------------------------------*/
{
  int i;
  char* nl;
  char* start;
  int line_number;

  start = DbLexLineStart ();
  line_number = DbLexLineNumber ();

  nl = (char*) strchr (start, '\n');
  if (nl != NULL) *nl = 0;

  DbSetErrorInfo (DbErrorBadSyntax, 
                  "Db> %s at line %d :\nDb> %s\n", s, line_number, start);
  DbPrintError ();

  if (nl != NULL) *nl = '\n';

  return (1);
}

/*---------------------------------------------------------------*/
static DbObject DbObjectTextAnalyze (ParserMode mode, char* text, CLList list)
/*---------------------------------------------------------------*/
{
  ParserContext context;
  ParserContext* previous_context = Context;
  LexContext* previous_lex_context;
  LexContext lex_context;
  DbObject result = NULL;

  Context = &context;

  Context->mode          = mode;
  Context->first_object  = NULL;
  Context->object        = NULL;
  Context->type          = NULL;
  Context->type_iterator = NULL;
  Context->objects       = list;
  Context->name          = NULL;

  previous_lex_context = DbLexStartup (&lex_context, text);

  DbSemanticAnalyzer ();
  DbLexRestoreContext (previous_lex_context);

  if (Context->name != NULL)
    {
      free (Context->name);
      Context->name = NULL;
    }

  result = Context->first_object;

  Context = previous_context;

  return (result);
}

/*---------------------------------------------------------------*/
DbObject DbParserCreateObject (char* text)
/*---------------------------------------------------------------*/
{
  DbObject object;

  object = DbObjectTextAnalyze (ObjectHead, text, NULL);

  return (object);
}

/*---------------------------------------------------------------*/
void DbParserCreateObjects (char* text, CLList list)
/*---------------------------------------------------------------*/
{
  DbObjectTextAnalyze (ObjectHead, text, list);
}

/*---------------------------------------------------------------*/
DbObject DbParserLoadObjectCopy (char* text)
/*---------------------------------------------------------------*/
{
  DbObject object;

  object = DbObjectTextAnalyze (ObjectCopy, text, NULL);

  return (object);
}

/*---------------------------------------------------------------*/
void DbParserLoadObjectCopies (char* text, CLList list)
/*---------------------------------------------------------------*/
{
  DbObjectTextAnalyze (ObjectCopy, text, list);
}

/*---------------------------------------------------------------*/
static void ContextSetName (ParserContext* context, char* name)
/*---------------------------------------------------------------*/
{
  if (context->name != NULL) 
    {
      free (context->name);
      context->name = NULL;
    }

  if (name == NULL) return;
  if (strlen (name) == 0) return;
  
  context->name = (char*) malloc (strlen (name) + 1);
  strcpy (context->name, name);
}

/*---------------------------------------------------------------*/
static void ContextAppendName (ParserContext* context, char* name)
/*---------------------------------------------------------------*/
{
  if (name == NULL) return;
  if (strlen (name) == 0) return;
  
  context->name = (char*) realloc (
      context->name, 
      strlen (context->name) + 1 + strlen (name) + 1);
  strcat (context->name, "/");
  strcat (context->name, name);
}

