#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <Db.h>

/*---------------------------------------------------------------------------*/
/*                                                                           */
/*  Visitor to build a string        	                                     */
/*                                                                           */
/*---------------------------------------------------------------------------*/

typedef struct _FormatterRec* Formatter;
typedef struct _FormatterRec
{
  DbVisitorRec base;
  char* string;
  int allocated;
  int length;
  int tab;
} FormatterRec;

/*---------------------------------------------------------------------------*/
static void append (Formatter visitor, char* text)
/*---------------------------------------------------------------------------*/
{
  int length;

  if (text == NULL) return;
  length = strlen (text);
  if (length == 0) return;

  while ((visitor->length + length + 1) > visitor->allocated)
    {
      visitor->allocated += 1000;

      if (visitor->string == NULL)
        {
          visitor->string = (char*) malloc (visitor->allocated);
          visitor->string[0] = 0;
        }
      else
        {
          visitor->string = (char*) realloc (visitor->string,
                                             visitor->allocated);
        }
    }

  strcat (visitor->string, text);
  visitor->length += length;
}

/*---------------------------------------------------------------------------*/
static void tab (Formatter visitor)
/*---------------------------------------------------------------------------*/
{
  int i;

  for (i = 0; i < visitor->tab; i++)
    {
      append (visitor, "  ");
    }
}

/*---------------------------------------------------------------------------*/
static int down (Formatter visitor, DbTypeIterator it)
/*---------------------------------------------------------------------------*/
{
  char* key;

  if (visitor == NULL) return (0);

  tab (visitor);

  key = DbTypeIteratorGetKey (it);
  if (key != NULL)
    {
      append (visitor, key);
      append (visitor, " =");
    }

  append (visitor, "\n");
  tab (visitor);

  append (visitor, "{\n");
  visitor->tab++;

  return (1);
}

/*---------------------------------------------------------------------------*/
static int in (Formatter visitor, DbTypeIterator it)
/*---------------------------------------------------------------------------*/
{
  void* address;

  if (visitor == NULL) return (0);

  address = DbTypeIteratorGetItemAddress (it);
  if (address != NULL)
    {
      DbType type;
      char* text;
      char* key;

      type    = DbTypeIteratorGetItemType (it);
      
      text = DbTypeOutput (type, address);

      if (text != NULL)
        {
          tab (visitor);

          key = DbTypeIteratorGetKey (it);
          if (key != NULL)
            {
              append (visitor, key);
              append (visitor, " = ");
            }

          append (visitor, text);
          append (visitor, "\n");
        }
    }

  return (1);
}

/*---------------------------------------------------------------------------*/
static int up (Formatter visitor, DbTypeIterator it)
/*---------------------------------------------------------------------------*/
{
  if (visitor == NULL) return (0);

  visitor->tab--;
  tab (visitor);
  append (visitor, "}\n");

  return (1);
}

/*---------------------------------------------------------------------------*/
static Formatter FormatterNew ()
/*---------------------------------------------------------------------------*/
{
  Formatter visitor = NULL;

  visitor = (Formatter) malloc (sizeof (FormatterRec));

  DbVisitorInit ((DbVisitor) visitor,
                 (DbVisitorFunction) NULL, 
                 (DbVisitorFunction) down, 
                 (DbVisitorFunction) in, 
                 (DbVisitorFunction) up);

  visitor->string    = NULL;
  visitor->allocated = 0;
  visitor->length    = 0;

  return (visitor);
}

/*---------------------------------------------------------------------------*/
static void FormatterReset (Formatter visitor, DbObject object)
/*---------------------------------------------------------------------------*/
{
  DbVersioningOption versioning;
  char* type_name;
  char* name;
  char* path_name;
  char* owner;

  if (visitor == NULL) return;
  if (object == NULL) return;

  visitor->length    = 0;
  if (visitor->string != NULL) visitor->string[0] = 0;
  visitor->tab = 0;

  versioning = DbObjectGetVersioningOption (object);
  switch (versioning)
    {
      case DbDefaultVersioning :
        break;
      case DbUseVersioning :
        append (visitor, "(version)");
        append (visitor, " ");
        break;
      case DbNoVersioning :
        append (visitor, "(noversion)");
        append (visitor, " ");
        break;
    }

  type_name = DbTypeName (DbObjectGetType (object));
  name = DbObjectGetName (object);
  path_name = DbObjectGetPathName (object);
  if (path_name == NULL) path_name = "";

  append (visitor, type_name);
  append (visitor, " ");
  if (strlen (path_name) > 0)
    {
      append (visitor, path_name);
      append (visitor, "/");
    }
  append (visitor, name);
  owner = DbObjectGetOwner (object);
  if (owner != NULL)
    {
      append (visitor, " @ ");
      append (visitor, owner);
    }
  append (visitor, " = ");
}

/*---------------------------------------------------------------*/
char* DbConvertObjectToString (DbObject object)
/*---------------------------------------------------------------*/
{
  static Formatter visitor = NULL;
  char* result;

  if (object == NULL) return (NULL);

  if (visitor == NULL)
    {
      visitor = FormatterNew ();
    }

  FormatterReset (visitor, object);

  DbObjectVisit (object, (DbVisitor) visitor);

  result = visitor->string;

  return (result);
}

/*---------------------------------------------------------------------*/
char* DbTypeConvertToString (DbType type)
/*---------------------------------------------------------------------*/
{
  static Formatter visitor = NULL;
  DbTStructure structure;
  CIterator it;
  char* name;

  if (!DbTypeIsStructure (type)) return (NULL);

  if (visitor == NULL)
    {
      visitor = FormatterNew ();
    }

  structure = (DbTStructure) type;
  name = DbTypeName (type);

  visitor->length    = 0;
  if (visitor->string != NULL) visitor->string[0] = 0;
  visitor->tab = 0;

  append (visitor, "structure ");
  append (visitor, name);
  append (visitor, " {\n");
  visitor->tab++;

  it = DbTStructureGiveFieldIterator (structure);
      
  for (; !CIteratorFinished (it); CIteratorForth (it))
    {
      DbField field;
      DbType type;
      char* type_name;
      
      field = (DbField) CIteratorItem (it);
      
      type = DbFieldType (field);
      
      type_name = DbTypeName (type);
      
      tab (visitor);
      append (visitor, type_name);
      append (visitor, " ");
      append (visitor, DbFieldGetName (field));
      append (visitor, "\n");
    }
  
  CIteratorDelete (it);

  visitor->tab--;
  append (visitor, "}");
  
  return (visitor->string);
}


